from relationalai.early_access.lqp import ir as lqp
from relationalai.early_access.lqp.utils import UniqueNames

rel_to_lqp = {
    "+": "rel_primitive_add",
    "-": "rel_primitive_subtract",
    "*": "rel_primitive_multiply",
    "/": "rel_primitive_divide",
    "=": "rel_primitive_eq",
    "!=": "rel_primitive_neq",
    "<=": "rel_primitive_lt_eq",
    ">=": "rel_primitive_gt_eq",
    ">": "rel_primitive_gt",
    "<": "rel_primitive_lt",
    "abs": "rel_primitive_abs",
    "construct_date": "rel_primitive_construct_date",
    "construct_datetime": "rel_primitive_construct_datetime",
    "construct_datetime_ms_tz": "rel_primitive_construct_datetime",
    "parse_date": "rel_primitive_parse_date",
    "parse_datetime": "rel_primitive_parse_datetime",
    "parse_decimal": "rel_primitive_parse_decimal",
    "parse_decimal64": "rel_primitive_parse_decimal",
    "parse_decimal128": "rel_primitive_parse_decimal",
    "starts_with": "rel_primitive_starts_with",
    "ends_with": "rel_primitive_ends_with",
    "contains": "rel_primitive_contains",
    "num_chars": "rel_primitive_num_chars",
    "substring": "rel_primitive_substring",
    "like_match": "rel_primitive_like_match",
    "concat": "rel_primitive_concat",
    "replace": "rel_primitive_replace",
    "date_year": "rel_primitive_date_year",
    "date_month": "rel_primitive_date_month",
    "date_day": "rel_primitive_date_day",
    "range": "rel_primitive_range",
}

agg_to_lqp = {
    "min": "rel_primitive_min",
    "max": "rel_primitive_max",
    "sum": "rel_primitive_add",
    "count": "rel_primitive_add", # count is a sum of 1s
    "rel_primitive_solverlib_ho_appl": "rel_primitive_solverlib_ho_appl",
}

def relname_to_lqp_name(name: str) -> str:
    # TODO: do these proprly
    if name in rel_to_lqp:
        return rel_to_lqp[name]
    else:
        # If we don't have a mapping for the built-in, we just pass it through as-is.
        return name

# We take the name and type of the variable that we're summing over, so that we can generate
# recognizable names for the variables in the reduce operation and preserve the type.
def lqp_avg_op(names: UniqueNames, op_name: str, sum_name: str, sum_type: lqp.RelType) -> lqp.Abstraction:
    count_type = lqp.PrimitiveType.INT
    vars = [
        (lqp.Var(name=names.get_name(sum_name), meta=None), sum_type),
        (lqp.Var(name=names.get_name("counter"), meta=None), count_type),
        (lqp.Var(name=names.get_name(sum_name), meta=None), sum_type),
        (lqp.Var(name=names.get_name("one"), meta=None), count_type),
        (lqp.Var(name=names.get_name("sum"), meta=None), sum_type),
        (lqp.Var(name=names.get_name("count"), meta=None), count_type),
    ]

    x1 = vars[0][0]
    x2 = vars[1][0]
    y1 = vars[2][0]
    y2 = vars[3][0]
    sum = vars[4][0]
    count = vars[5][0]

    body = lqp.Conjunction(
        args=[
            lqp.Primitive(name="rel_primitive_add", terms=[x1, y1, sum], meta=None),
            lqp.Primitive(name="rel_primitive_add", terms=[x2, y2, count], meta=None)
        ],
        meta=None
    )
    return lqp.Abstraction(vars=vars, value=body, meta=None)

# Default handler for aggregation operations in LQP.
def lqp_agg_op(names: UniqueNames, op_name: str, aggr_arg_name: str, aggr_arg_type: lqp.RelType) -> lqp.Abstraction:
    x = lqp.Var(name=names.get_name(f"x_{aggr_arg_name}"), meta=None)
    y = lqp.Var(name=names.get_name(f"y_{aggr_arg_name}"), meta=None)
    z = lqp.Var(name=names.get_name(f"z_{aggr_arg_name}"), meta=None)
    ts = [(x, aggr_arg_type), (y, aggr_arg_type), (z, aggr_arg_type)]

    name = agg_to_lqp.get(op_name, op_name)
    body = lqp.Primitive(name=name, terms=[x, y, z], meta=None)

    return lqp.Abstraction(vars=ts, value=body, meta=None)

def lqp_operator(names: UniqueNames, op_name: str, aggr_arg_name: str, aggr_arg_type: lqp.RelType) -> lqp.Abstraction:
    # TODO: Can we just pass through unknown operations?
    if op_name not in agg_to_lqp:
        raise NotImplementedError(f"Unsupported aggregation: {op_name}")

    return lqp_agg_op(names, op_name, aggr_arg_name, aggr_arg_type)



import relationalai.early_access.builder as qb


def reference_entity(concept: qb.Concept, *args):
    kwargs = entity_kwargs(concept, *args, shallow=True)
    return concept.new_identity(**kwargs)

def construct_entity(concept: qb.Concept, *args):
    kwargs = entity_kwargs(concept, *args)
    return concept.new(**kwargs)

def entity_kwargs(concept: qb.Concept, *args, shallow=False):
    ref_scheme = concept._ref_scheme(shallow=shallow)
    if not ref_scheme:
        raise ValueError(f'Concept {concept} has no reference scheme defined')
    keys = [rel._short_name for rel in ref_scheme]
    if len(args) != len(keys):
        raise ValueError(f'Expected {len(keys)} arguments for constructor of {concept}, got {len(args)}')
    kwargs = dict(zip(keys, args))
    return kwargs



"""
Framework for performing dataflow analysis on the IR.
"""
from dataclasses import dataclass, field
from typing import Dict, TypeVar, Generic
from functools import reduce

from .util import OrderedSet, ordered_set
from . import ir

T = TypeVar('T')

@dataclass
class DataflowAnalyzer(Generic[T]):
    """
    A framework for performing dataflow analysis on the IR.

    The analysis works over a semilattice `T`, which is a set of values that can be combined
    using a meet() operation. The meet() operation is must be:
    - Associative: (a ∧ b) ∧ c = a ∧ (b ∧ c)
    - Commutative: a ∧ b = b ∧ a
    - Idempotent: a ∧ a = a

    Subclasses must also specify a bottom() element for `T`.

    Typically elements of `T` are a set and the meet operation is intersection or union.
    Users can overload transfer() to define their specific analysis.
    """
    def __init__(self):
        # Maps nodes to their dataflow values
        self._in: Dict[ir.Node, T] = field(default_factory=dict)
        self._out: Dict[ir.Node, T] = field(default_factory=dict)
        self._succs: Dict[ir.Node, OrderedSet[ir.Node]] = field(default_factory=dict)
        self._preds: Dict[ir.Node, OrderedSet[ir.Node]] = field(default_factory=dict)

    def transfer(self, node: ir.Node, in_val: T) -> T:
        """
        Transfer function that defines how data flows through a node.
        Override this to define the analysis-specific transfer function.
        """
        return in_val

    def meet(self, a: T, b: T) -> T:
        """
        Meet operation that combines two dataflow values.
        """
        raise NotImplementedError

    def add_flow(self, src: ir.Node, dst: ir.Node):
        """Record a flow from src to dst."""
        self._succs[src].add(dst)
        self._preds[dst].add(src)

    def bottom(self) -> T:
        """Create a bottom element for the semilattice."""
        raise NotImplementedError

    def _get_predecessors(self, node: ir.Node) -> OrderedSet[ir.Node]:
        """Get all nodes that flow into this node."""
        return self._preds.get(node, ordered_set())

    def _get_successors(self, node: ir.Node) -> OrderedSet[ir.Node]:
        """Get all nodes that this node flows into."""
        return self._succs.get(node, ordered_set())

    def _solve(self):
        """Solve the dataflow equations using a worklist algorithm."""
        # Initialize worklist with all nodes that have flows
        worklist = ordered_set()
        for src in self._succs:
            worklist.add(src)

        while worklist:
            node = next(iter(worklist))
            worklist.remove(node)

            # Get predecessors' out values
            pred_outs = [
                self._out.get(pred, self.bottom()) 
                for pred in self._get_predecessors(node)
            ]
            
            # Compute new in value using meet
            new_in = reduce(self.meet, pred_outs, self.bottom())
            # Compute new out value using transfer
            new_out = self.transfer(node, new_in)

            # Update values if they changed
            if (new_in != self._in.get(node, self.bottom()) or 
                new_out != self._out.get(node, self.bottom())):
                self._in[node] = new_in
                self._out[node] = new_out
                # Add successors to worklist
                worklist.update(self._get_successors(node))

    def analyze(self, root: ir.Node):
        """
        Build the flow graph and solve the dataflow equations.
        This should be called after all flows have been added.
        """
        self._solve()

    def get_in(self, node: ir.Node) -> T:
        """Get the dataflow value flowing into a node."""
        return self._in.get(node, self.bottom())

    def get_out(self, node: ir.Node) -> T:
        """Get the dataflow value flowing out of a node."""
        return self._out.get(node, self.bottom())


from relationalai.early_access.dsl.bindings.common import Binding, BindableTable
from relationalai.early_access.dsl.orm.constraints import Unique
from relationalai.early_access.dsl.orm.relationships import Relationship, Role
from relationalai.early_access.dsl.orm.types import Concept


class BoundRelationship:
    _relationship: Relationship
    _table: BindableTable
    _bindings: list[Binding]

    def __init__(self, relationship: Relationship, table: BindableTable):
        self._relationship = relationship
        self._table = table
        self._bindings = []

    def __hash__(self):
        # hash based on the relationship and the table
        return hash((self._relationship, self._table))

    def __eq__(self, other):
        if not isinstance(other, BoundRelationship):
            return False
        return self._relationship == other._relationship and self._table == other._table

    @property
    def relationship(self):
        return self._relationship

    @property
    def table(self):
        return self._table

    @property
    def bindings(self):
        return self._bindings

class BoundIdentifierConstraint:
    _constraint: Unique
    _table: BindableTable
    _role_bindings: dict[Role, Binding]
    _concept: Concept

    def __init__(self, constraint: Unique, table: BindableTable, role_bindings: dict[Role, Binding]):
        self._constraint = constraint
        self._table = table
        self._role_bindings = role_bindings
        self._lookup_value_concept(constraint)

    def _lookup_value_concept(self, constraint):
        # concept if the sibling player of an arbitrary role (all should have the same) in the constraint
        first_role = constraint.roles()[0]
        sibling_role = first_role.sibling()
        if sibling_role is None:
            raise ValueError('Roles in external UCs must be part of binary relationships')
        self._concept = sibling_role.player()

    def __hash__(self):
        # hash based on the constraint and the table
        return hash((self._constraint, self._table))

    @property
    def constraint(self):
        return self._constraint

    @property
    def table(self):
        return self._table

    @property
    def role_bindings(self):
        return self._role_bindings

    @property
    def concept(self):
        return self._concept

    def is_external(self):
        return len(self._role_bindings) > 1



"""
Core functionality for the graphlib package.
"""
from typing import Optional

from relationalai.early_access.builder import Concept, Relationship
from relationalai.early_access.builder import Integer, Float
from relationalai.early_access.builder import where, define, count, sum, not_, min, union
from relationalai.early_access.builder import avg
from relationalai.early_access.builder.std.math import abs, natural_log

class Graph():
    def __init__(self,
            *,
            directed: bool,
            weighted: bool,
            aggregator: Optional[str] = None,
        ):
        assert isinstance(directed, bool), "The `directed` argument must be a boolean."
        assert isinstance(weighted, bool), "The `weighted` argument must be a boolean."
        self.directed = directed
        self.weighted = weighted

        assert isinstance(aggregator, type(None)), "Weight aggregation not yet supported."
        # TODO: In the hopefully not-too-distant future, this argument will
        #   allow the user to specify whether and how to aggregate weights
        #   for multi-edges that exist at the user interface (Edge) level
        #   to construct the internal edge/weight list representation.
        #   The `str` type is just a placeholder; it should be something else.

        # Introduce Node and Edge concepts.
        Node = Concept("Node")
        Edge = Concept("Edge")
        Edge.src = Relationship("{edge:Edge} has source {src:Node}")
        Edge.dst = Relationship("{edge:Edge} has destination {dst:Node}")
        Edge.weight = Relationship("{edge:Edge} has weight {weight:Float}")
        self.Node = Node
        self.Edge = Edge

        # TODO: Require that each Edge has an Edge.src.
        # TODO: Require that each Edge has an Edge.dst.
        # TODO: If weighted, require that each Edge has an Edge.weight.
        # TODO: If not weighted, require that each Edge does not have an Edge.weight.

        # TODO: Suppose that type checking should in future restrict `src` and
        #   `dst` to be `Node`s, but at the moment we may need a require for that.
        # TODO: Suppose that type checking should in future restrict `weight` to be
        #   `Float`s, but at the moment we may need a require for that.

        # TODO: Transform Node and Edge into underlying edge-/weight-list representation.
        # NOTE: Operate under the assumption that `Node` contains all
        #   possible nodes, i.e. we can use the `Node` Concept directly as
        #   the node list. Has the additional benefit of allowing relationships
        #   (for which it makes sense) to be properties of `Node` rather than standalone.
        self._define_edge_relationships()
 
        self._define_num_nodes_relationship()
        self._define_num_edges_relationship()

        self._define_neighbor_relationships()
        self._define_count_neighbor_relationships()
        self._define_common_neighbor_relationship()
        self._define_count_common_neighbor_relationship()

        self._define_degree_relationships()
        self._define_weighted_degree_relationships()

        self._define_degree_centrality_relationship()

        self._define_reachable_from()

        # Helper relationship for preferential attachment.
        self._define_isolated_node_relationship()

        self._define_preferential_attachment_relationship()

        # Helper relationships for triangle functionality.
        self._define_no_loop_edge_relationship()
        self._define_oriented_edge_relationship()
        self._define_reversed_oriented_edge_relationship()

        self._define_triangle_count_relationship()
        self._define_unique_triangle_relationship()
        self._define_num_triangles_relationship()
        self._define_triangle_relationship()

        # Helper relationships for local clustering coefficient.
        self._define_degree_no_self_relationship()

        self._define_local_clustering_coefficient_relationship()
        self._define_average_clustering_coefficient_relationship()
        self._define_adamic_adar_relationship()

        self._define_weakly_connected_component()

        self._define_distance_relationship()

    def _define_edge_relationships(self):
        """
        Define the self._edge and self._weight relationships,
        consuming the Edge concept's `src`, `dst`, and `weight` relationships.
        """
        self._edge = Relationship("{src:Node} has edge to {dst:Node}")
        self._weight = Relationship("{src:Node} has edge to {dst:Node} with weight {weight:Float}")

        Edge = self.Edge
        if self.directed and self.weighted:
            define(
                self._weight(Edge.src, Edge.dst, Edge.weight),
                self._edge(Edge.src, Edge.dst)
            )
        elif self.directed and not self.weighted:
            define(
                self._weight(Edge.src, Edge.dst, 1.0),
                self._edge(Edge.src, Edge.dst)
            )
        elif not self.directed and self.weighted:
            define(
                self._weight(Edge.src, Edge.dst, Edge.weight),
                self._weight(Edge.dst, Edge.src, Edge.weight),
                self._edge(Edge.src, Edge.dst),
                self._edge(Edge.dst, Edge.src)
            )
        elif not self.directed and not self.weighted:
            define(
                self._weight(Edge.src, Edge.dst, 1.0),
                self._weight(Edge.dst, Edge.src, 1.0),
                self._edge(Edge.src, Edge.dst),
                self._edge(Edge.dst, Edge.src)
            )

    def _define_num_nodes_relationship(self):
        """Define the self._num_nodes relationship."""
        self._num_nodes = Relationship("The graph has {num_nodes:Integer} nodes")
        define(self._num_nodes(count(self.Node) | 0))

    def _define_num_edges_relationship(self):
        """Define the self._num_edges relationship."""
        self._num_edges = Relationship("The graph has {num_edges:Integer} edges")

        src, dst = self.Node.ref(), self.Node.ref()

        if self.directed:
            define(self._num_edges(count(src, dst, self._edge(src, dst)) | 0))
        elif not self.directed:
            define(self._num_edges(count(src, dst, self._edge(src, dst), src <= dst) | 0))
            # TODO: Generates an UnresolvedOverload warning from the typer.
            #   Should be sorted out by improvements in the typer (to allow
            #   comparisons between instances of concepts).


    def _define_neighbor_relationships(self):
        """Define the self.[in,out]neighbor relationships."""
        self._neighbor = Relationship("{src:Node} has neighbor {dst:Node}")
        self._inneighbor = Relationship("{dst:Node} has inneighbor {src:Node}")
        self._outneighbor = Relationship("{src:Node} has outneighbor {dst:Node}")

        src, dst = self.Node.ref(), self.Node.ref()
        where(self._edge(src, dst)).define(self._neighbor(src, dst), self._neighbor(dst, src))
        where(self._edge(dst, src)).define(self._inneighbor(src, dst))
        where(self._edge(src, dst)).define(self._outneighbor(src, dst))
        # Note that these definitions happen to work for both
        # directed and undirected graphs due to `edge` containing
        # each edge's symmetric partner in the undirected case.

    def _define_count_neighbor_relationships(self):
        """
        Define the self._count_[in,out]neighbor relationships.
        Note that these relationships differ from corresponding
        [in,out]degree relationships in that they yield empty
        rather than zero absent [in,out]neighbors.
        Primarily for internal consumption.
        """
        self._count_neighbor = Relationship("{src:Node} has neighbor count {count:Integer}")
        self._count_inneighbor = Relationship("{dst:Node} has inneighbor count {count:Integer}")
        self._count_outneighbor = Relationship("{src:Node} has outneighbor count {count:Integer}")

        src, dst = self.Node.ref(), self.Node.ref()
        where(self._neighbor(src, dst)).define(self._count_neighbor(src, count(dst).per(src)))
        where(self._inneighbor(dst, src)).define(self._count_inneighbor(dst, count(src).per(dst)))
        where(self._outneighbor(src, dst)).define(self._count_outneighbor(src, count(dst).per(src)))


    def _define_common_neighbor_relationship(self):
        """Define the self._common_neighbor relationship."""
        self._common_neighbor = Relationship("{node_a:Node} and {node_b:Node} have common neighbor {node_c:Node}")

        node_a, node_b, node_c = self.Node.ref(), self.Node.ref(), self.Node.ref()
        where(self._neighbor(node_a, node_c), self._neighbor(node_b, node_c)).define(self._common_neighbor(node_a, node_b, node_c))

    def _define_count_common_neighbor_relationship(self):
        """Define the self._count_common_neighbor relationship."""
        self._count_common_neighbor = Relationship("{node_a:Node} and {node_b:Node} have common neighbor count {count:Integer}")

        node_a, node_b, node_c = self.Node.ref(), self.Node.ref(), self.Node.ref()
        where(self._common_neighbor(node_a, node_b, node_c)).define(self._count_common_neighbor(node_a, node_b, count(node_c).per(node_a, node_b)))


    def _define_degree_relationships(self):
        """Define the self._[in,out]degree relationships."""
        self._degree = Relationship("{node:Node} has degree {count:Integer}")
        self._indegree = Relationship("{node:Node} has indegree {count:Integer}")
        self._outdegree = Relationship("{node:Node} has outdegree {count:Integer}")

        incount, outcount = Integer.ref(), Integer.ref()

        where(
            self.Node,
            _indegree := where(self._count_inneighbor(self.Node, incount)).select(incount) | 0,
        ).define(self._indegree(self.Node, _indegree))

        where(
            self.Node,
            _outdegree := where(self._count_outneighbor(self.Node, outcount)).select(outcount) | 0,
        ).define(self._outdegree(self.Node, _outdegree))

        if self.directed:
            where(
                self._indegree(self.Node, incount),
                self._outdegree(self.Node, outcount),
            ).define(self._degree(self.Node, incount + outcount))
        elif not self.directed:
            neighcount = Integer.ref()
            where(
                self.Node,
                _degree := where(self._count_neighbor(self.Node, neighcount)).select(neighcount) | 0,
            ).define(self._degree(self.Node, _degree))

    def _define_reachable_from(self):
        """Define the self.reachable_from relationship"""
        self._reachable_from = Relationship("{node_a:Node} reaches {node_b:Node}")

        node_a, node_b, node_c = self.Node.ref(), self.Node.ref(), self.Node.ref()
        define(self._reachable_from(node_a, node_a))
        define(self._reachable_from(node_a, node_c)).where(self._reachable_from(node_a, node_b), self._edge(node_b, node_c))


    def _define_weighted_degree_relationships(self):
        """Define the self._weighted_[in,out]degree relationships."""
        self._weighted_degree = Relationship("{node:Node} has weighted degree {weight:Float}")
        self._weighted_indegree = Relationship("{node:Node} has weighted indegree {weight:Float}")
        self._weighted_outdegree = Relationship("{node:Node} has weighted outdegree {weight:Float}")

        src, dst = self.Node.ref(), self.Node.ref()
        inweight, outweight = Float.ref(), Float.ref()

        where(
            self.Node,
            _weighted_indegree := sum(src, inweight).per(self.Node).where(self._weight(src, self.Node, inweight)) | 0.0,
        ).define(self._weighted_indegree(self.Node, _weighted_indegree))

        where(
            self.Node,
            _weighted_outdegree := sum(dst, outweight).per(self.Node).where(self._weight(self.Node, dst, outweight)) | 0.0,
        ).define(self._weighted_outdegree(self.Node, _weighted_outdegree))

        if self.directed:
            where(
                self._weighted_indegree(self.Node, inweight),
                self._weighted_outdegree(self.Node, outweight),
            ).define(self._weighted_degree(self.Node, inweight + outweight))
        elif not self.directed:
            weight = Float.ref()
            where(
                self.Node,
                _weighted_degree := sum(dst, weight).per(self.Node).where(self._weight(self.Node, dst, weight)) | 0.0,
            ).define(self._weighted_degree(self.Node, _weighted_degree))


    def _define_degree_centrality_relationship(self):
        """Define the self._degree_centrality relationship."""
        self._degree_centrality = Relationship("{node:Node} has {degree_centrality:Float}")

        degree = Integer.ref()
        weighted_degree = Float.ref()

        # A single isolated node has degree centrality zero.
        where(
            self._num_nodes(1),
            self._degree(self.Node, 0)
        ).define(self._degree_centrality(self.Node, 0.0))

        # A single non-isolated node has degree centrality one.
        where(
            self._num_nodes(1),
            self._degree(self.Node, degree),
            degree > 0
        ).define(self._degree_centrality(self.Node, 1.0))

        # General case, i.e. with more than one node.
        num_nodes = Integer.ref()
        if self.weighted:
            where(
                self._num_nodes(num_nodes),
                num_nodes > 1,
                self._weighted_degree(self.Node, weighted_degree)
            ).define(self._degree_centrality(self.Node, weighted_degree / (num_nodes - 1.0)))
        elif not self.weighted:
            where(
                self._num_nodes(num_nodes),
                num_nodes > 1,
                self._degree(self.Node, degree)
            ).define(self._degree_centrality(self.Node, degree / (num_nodes - 1.0)))


    def _define_isolated_node_relationship(self):
        """Define the self._isolated_node (helper, non-public) relationship."""
        self._isolated_node = Relationship("{node:Node} is isolated")

        dst = self.Node.ref()
        where(
            self.Node,
            not_(self._neighbor(self.Node, dst))
        ).define(self._isolated_node(self.Node))

    def _define_preferential_attachment_relationship(self):
        """Define the self._preferential_attachment relationship."""
        self._preferential_attachment = Relationship("{node_u:Node} and {node_v:Node} have preferential attachment score {score:Integer}")

        node_u, node_v = self.Node.ref(), self.Node.ref()
        count_u, count_v = Integer.ref(), Integer.ref()

        # NOTE: We consider isolated nodes separately to maintain
        #   the dense behavior of preferential attachment.

        # Case where node u is isolated, and node v is any node: score 0.
        where(
            self._isolated_node(node_u),
            self.Node(node_v),
        ).define(self._preferential_attachment(node_u, node_v, 0))

        # Case where node u is any node, and node v is isolated: score 0.
        where(
            self.Node(node_u),
            self._isolated_node(node_v)
        ).define(self._preferential_attachment(node_u, node_v, 0))

        # Case where neither node is isolated: score is count_neighbor[u] * count_neighbor[v].
        where(
            self._count_neighbor(node_u, count_u),
            self._count_neighbor(node_v, count_v)
        ).define(self._preferential_attachment(node_u, node_v, count_u * count_v))

    def _define_weakly_connected_component(self):
        """Defines the self.weakly_connected_component relationship"""
        self._weakly_connected_component = Relationship("{node:Node} is in the connected component {id:Node}")

        node, node_v, component = self.Node.ref(), self.Node.ref(), self.Node.ref()
        node, component = union(
            # A node starts with itself as the component id.
            where(node == component).select(node, component),
            # Recursive case.
            where(self._weakly_connected_component(node, component), self._neighbor(node, node_v)).select(node_v, component)
        )
        define(self._weakly_connected_component(node, min(component).per(node)))

    def _define_no_loop_edge_relationship(self):
        """Define the self._no_loop_edge (helper, non-public) relationship."""
        self._no_loop_edge = Relationship("{src:Node} has nonloop edge to {dst:Node}")

        src, dst = self.Node.ref(), self.Node.ref()
        where(
            self._edge(src, dst),
            src != dst
        ).define(self._no_loop_edge(src, dst))

    def _define_oriented_edge_relationship(self):
        """Define the self._oriented_edge (helper, non-public) relationship."""
        self._oriented_edge = Relationship("{src:Node} has oriented edge to {dst:Node}")

        src, dst = self.Node.ref(), self.Node.ref()
        where(
            self._edge(src, dst),
            src < dst
        ).define(self._oriented_edge(src, dst))

    def _define_reversed_oriented_edge_relationship(self):
        """Define the self._reversed_oriented_edge (helper, non-public) relationship."""
        self._reversed_oriented_edge = Relationship("{src:Node} has reversed oriented edge to {dst:Node}")

        src, dst = self.Node.ref(), self.Node.ref()
        where(
            self._edge(src, dst),
            src > dst
        ).define(self._reversed_oriented_edge(src, dst))


    def _define_triangle_count_relationship(self):
        """Define self._triangle_count relationship."""
        self._triangle_count = Relationship("{node:Node} belongs to {count:Integer} triangles")

        where(
            self.Node,
            _count := self._nonzero_triangle_count_fragment(self.Node) | 0
        ).define(self._triangle_count(self.Node, _count))

    def _nonzero_triangle_count_fragment(self, node):
        """
        Helper function that returns a fragment, specifically a count
        of the number of triangles containing the given node.
        """
        node_a, node_b = self.Node.ref(), self.Node.ref()

        if self.directed:
            # For directed graphs, count triangles with any circulation.
            # For example, count both (1-2-3-1) and (1-3-2-1) as triangles.
            return count(node_a, node_b).per(node).where(
                self._no_loop_edge(node, node_a),
                self._no_loop_edge(node_a, node_b),
                self._no_loop_edge(node_b, node)
            )
        else:
            # For undirected graphs, count triangles with a specific circulation.
            # For example, count (1-2-3-1) but not (1-3-2-1) as a triangle.
            return count(node_a, node_b).per(node).where(
                self._no_loop_edge(node, node_a),
                self._oriented_edge(node_a, node_b),
                self._no_loop_edge(node, node_b)
            )


    def _define_unique_triangle_relationship(self):
        """Define self._unique_triangle relationship."""
        self._unique_triangle = Relationship("{node_a:Node} and {node_b:Node} and {node_c:Node} form unique triangle")

        node_a, node_b, node_c = self.Node.ref(), self.Node.ref(), self.Node.ref()

        where(
            self._unique_triangle_fragment(node_a, node_b, node_c)
        ).define(self._unique_triangle(node_a, node_b, node_c))

    def _unique_triangle_fragment(self, node_a, node_b, node_c):
        """
        Helper function that returns a fragment, specifically a where clause
        constraining the given triplet of nodes to unique triangles in the graph.
        """
        if self.directed:
            return where(
                self._oriented_edge(node_a, node_b),
                self._no_loop_edge(node_b, node_c),
                self._reversed_oriented_edge(node_c, node_a)
            )
        else:
            return where(
                self._oriented_edge(node_a, node_b),
                self._oriented_edge(node_b, node_c),
                self._oriented_edge(node_a, node_c)
            )


    def _define_num_triangles_relationship(self):
        """Define self._num_triangles relationship."""
        self._num_triangles = Relationship("The graph has {num_triangles:Integer} triangles")

        _num_triangles = Integer.ref()
        node_a, node_b, node_c = self.Node.ref(), self.Node.ref(), self.Node.ref()

        where(
            _num_triangles := count(
                node_a, node_b, node_c
            ).where(
                self._unique_triangle_fragment(node_a, node_b, node_c)
            ) | 0,
        ).define(self._num_triangles(_num_triangles))

    def _define_triangle_relationship(self):
        """Define self._triangle relationship."""
        self._triangle = Relationship("{node_a:Node} and {node_b:Node} and {node_c:Node} form a triangle")

        a, b, c = self.Node.ref(), self.Node.ref(), self.Node.ref()

        if self.directed:
            where(self._unique_triangle(a, b, c)).define(self._triangle(a, b, c))
            where(self._unique_triangle(b, c, a)).define(self._triangle(a, b, c))
            where(self._unique_triangle(c, a, b)).define(self._triangle(a, b, c))
        else:
            where(self._unique_triangle(a, b, c)).define(self._triangle(a, b, c))
            where(self._unique_triangle(a, c, b)).define(self._triangle(a, b, c))
            where(self._unique_triangle(b, a, c)).define(self._triangle(a, b, c))
            where(self._unique_triangle(b, c, a)).define(self._triangle(a, b, c))
            where(self._unique_triangle(c, a, b)).define(self._triangle(a, b, c))
            where(self._unique_triangle(c, b, a)).define(self._triangle(a, b, c))


    def _define_degree_no_self_relationship(self):
        """
        Define self._degree_no_self relationship
        (non-public helper for local clustering coefficient).
        """
        self._degree_no_self = Relationship("{node:Node} has degree excluding self loops {num:Integer}")

        node, neighbor = self.Node.ref(), self.Node.ref()
        where(
            self.Node(node),
            _dns := count(neighbor).per(node).where(self._no_loop_edge(node, neighbor)) | 0,
        ).define(self._degree_no_self(node, _dns))

    def _define_local_clustering_coefficient_relationship(self):
        """
        Define self._local_clustering_coefficient relationship.
        Note that local_clustering_coefficient only applies to undirected graphs.
        """
        self._local_clustering_coefficient = Relationship("{node:Node} has local clustering coefficient {coefficient:Float}")

        if self.directed:
            return

        node = self.Node.ref()
        degree_no_self = Integer.ref()
        triangle_count = Integer.ref()
        where(
            node,
            _lcc := where(
                self._degree_no_self(node, degree_no_self),
                self._triangle_count(node, triangle_count),
                degree_no_self > 1
            ).select(
                2.0 * triangle_count / (degree_no_self * (degree_no_self - 1.0))
            ) | 0.0,
        ).define(self._local_clustering_coefficient(node, _lcc))

    def _define_average_clustering_coefficient_relationship(self):
        """
        Define self._average_clustering_coefficient relationship.
        Note that average_clustering_coefficient only applies to undirected graphs.
        """
        self._average_clustering_coefficient = Relationship("The graph has average clustering coefficient {coefficient:Float}")

        if self.directed:
            return

        node = self.Node.ref()
        coefficient = Float.ref()
        where(
            _avg_coefficient := avg(node, coefficient).where(
                    self._local_clustering_coefficient(node, coefficient)
                ) | 0.0
        ).define(self._average_clustering_coefficient(_avg_coefficient))

    def _define_distance_relationship(self):
        """Define self._distance relationship."""
        if not self.weighted:
            self._distance = Relationship("{node_u:Node} and {node_v:Node} have a distance of {d:Integer}")
            node_u, node_v, node_n, d1 = self.Node.ref(), self.Node.ref(), self.Node.ref(), Integer.ref()
            node_u, node_v, d = union(
                where(node_u == node_v, d1 == 0).select(node_u, node_v, d1), # Base case.
                where(self._edge(node_n, node_v),
                      d2 := self._distance(node_u, node_n, Integer) + 1).select(node_u, node_v, d2) # Recursive case.
            )
            define(self._distance(node_u, node_v, min(d).per(node_u, node_v)))
        else:
            self._distance = Relationship("{node_u:Node} and {node_v:Node} have a distance of {d:Float}")
            node_u, node_v, node_n, w, d1 = self.Node.ref(), self.Node.ref(),\
                self.Node.ref(), Float.ref(), Float.ref()
            node_u, node_v, d = union(
                where(node_u == node_v, d1 == 0.0).select(node_u, node_v, d1), # Base case.
                where(self._weight(node_n, node_v, w), d2 := self._distance(node_u, node_n, Float) + abs(w))\
                .select(node_u, node_v, d2) # Recursive case.
            )
            define(self._distance(node_u, node_v, min(d).per(node_u, node_v)))

    def _define_adamic_adar_relationship(self):
        """Define self._adamic_adar relationship."""
        self._adamic_adar = Relationship("{node_u:Node} and {node_v:Node} have adamic adar score {score:Float}")

        node_u, node_v, common_neighbor = self.Node.ref(), self.Node.ref(), self.Node.ref()
        neighbor_count = Integer.ref()

        where(
            _score := sum(common_neighbor, 1.0 / natural_log(neighbor_count)).per(node_u, node_v).where(
                self._common_neighbor(node_u, node_v, common_neighbor),
                self._count_neighbor(common_neighbor, neighbor_count),
            )
        ).define(self._adamic_adar(node_u, node_v, _score))


    # Public accessor methods for private relationships.
    def num_nodes(self): return self._num_nodes
    def num_edges(self): return self._num_edges

    def neighbor(self): return self._neighbor
    def inneighbor(self): return self._inneighbor
    def outneighbor(self): return self._outneighbor

    def count_neighbor(self): return self._count_neighbor
    def count_inneighbor(self): return self._count_inneighbor
    def count_outneighbor(self): return self._count_outneighbor

    def common_neighbor(self): return self._common_neighbor
    def count_common_neighbor(self): return self._count_common_neighbor

    def degree(self): return self._degree
    def indegree(self): return self._indegree
    def outdegree(self): return self._outdegree

    def weighted_degree(self): return self._weighted_degree
    def weighted_indegree(self): return self._weighted_indegree
    def weighted_outdegree(self): return self._weighted_outdegree

    def degree_centrality(self): return self._degree_centrality

    def reachable_from(self): return self._reachable_from

    def preferential_attachment(self): return self._preferential_attachment

    def triangle_count(self): return self._triangle_count
    def unique_triangle(self): return self._unique_triangle
    def num_triangles(self): return self._num_triangles
    def triangle(self): return self._triangle

    def local_clustering_coefficient(self):
        if self.directed:
            # TODO: Eventually make this error more similar to
            #   the corresponding error emitted from the pyrel graphlib wrapper.
            raise NotImplementedError(
                "Local clustering coefficient is not applicable to directed graphs"
            )
        return self._local_clustering_coefficient

    def average_clustering_coefficient(self):
        if self.directed:
            raise NotImplementedError(
                "Average clustering coefficient is not applicable to directed graphs"
            )
        return self._average_clustering_coefficient

    def weakly_connected_component(self): return self._weakly_connected_component

    def distance(self): return self._distance
    def adamic_adar(self): return self._adamic_adar



from relationalai.early_access.dsl.core.constraints.predicate.atomic import Atom
from relationalai.early_access.dsl.core.exprs import Expr, Wildcard


class Domain(Expr):

    def __init__(self, v):
        self.part = v

    def display(self):
        return f"dom {self.part.display()}"

    def grounded_using(self, groundings): return False

    def decorated(self) -> bool:
        return self.part.decorated()

    def entityid(self):
        return hash((Domain, self.part.entityid()))

    def project(self, comp):
        nargs = self.part.typeof().arity() - 1
        wc = Wildcard()

        seq = [wc] * nargs
        seq.insert(0, comp)

        return Atom(self.part, seq)

    def scalar_refs(self): return self.part.scalar_refs()

    def refersto(self, varname: str) -> bool:
        return self.part.refersto(varname)

    def relational(self) -> bool: return True

    def rename(self, renaming):
        return Domain(self.part.rename(renaming))

    def simplify(self): return self

    def substitute(self, bindings):
        p = self.part.substitute(bindings)
        return Domain(p)

    def undecorated(self) -> bool:
        return self.part.undecorated()


class Range(Expr):
    def __init__(self, v):
        self.part = v

    def display(self):
        return f"ran {self.part.display()}"

    def entityid(self):
        return hash((Range, self.part.entityid()))

    def grounded_using(self, groundings): return False

    def decorated(self) -> bool:
        return self.part.decorated()

    def project(self, comp):
        nargs = self.part.arity()
        seq = [Wildcard()] * nargs - 1
        seq.append(comp)

        return Atom(self.part, seq)

    def scalar_refs(self): return self.part.scalar_refs()

    def refersto(self, varname: str) -> bool:
        return self.part.refersto(varname)

    def relational(self) -> bool: return True

    def rename(self, renaming):
        return Range(self.part.rename(renaming))

    def simplify(self): return self

    def substitute(self, bindings):
        p = self.part.substitute(bindings)
        return Range(p)

    def undecorated(self) -> bool:
        return self.part.undecorated()



from relationalai.early_access.metamodel import ir, builtins as rel_builtins, helpers, types
from relationalai.early_access.metamodel.visitor import collect_by_type
from relationalai.early_access.lqp import ir as lqp
from relationalai.early_access.lqp.primitives import relname_to_lqp_name, lqp_operator, lqp_avg_op
from relationalai.early_access.lqp.types import meta_type_to_lqp, type_from_constant
from relationalai.early_access.lqp.constructors import mk_and, mk_exists, mk_or, mk_abstraction, mk_var
from relationalai.early_access.lqp.utils import TranslationCtx, gen_unique_var, rename_vars_formula, lqp_hash
from relationalai.early_access.lqp.validators import assert_valid_input

from typing import Tuple, cast, Union

""" Main access point. Converts the model IR to an LQP transaction. """
def to_lqp(model: ir.Model, fragment_name: bytes) -> lqp.Transaction:
    assert_valid_input(model)
    ctx = TranslationCtx(model)
    decls: list[lqp.Declaration] = []

    # LQP only accepts logical tasks
    # These are asserted at init time
    root = cast(ir.Logical, model.root)
    for subtask in root.body:
        assert isinstance(subtask, ir.Logical)
        new_decls = _translate_to_decls(ctx, subtask)
        decls.extend(new_decls)

    reads: list[lqp.Read] = []
    for (i, output_id) in enumerate(ctx.output_ids):
        assert isinstance(output_id, lqp.RelationId)
        output = lqp.Output(name=f"output_{i}", relation_id=output_id, meta=None)
        reads.append(lqp.Read(read_type=output, meta=None))

    debug_info = lqp.DebugInfo(id_to_orig_name=ctx.id_to_orig_name, meta=None)
    fragment_id = lqp.FragmentId(id=fragment_name, meta=None)
    fragment = lqp.Fragment(id=fragment_id, declarations=decls, meta=None, debug_info=debug_info)
    define_op = lqp.Define(fragment=fragment, meta=None)

    txn = lqp.Transaction(
        epochs=[
            lqp.Epoch(
                reads=reads,
                local_writes=[lqp.Write(write_type=define_op, meta=None)],
                persistent_writes=[],
                meta=None
            )
        ],
        meta=None,
    )

    lqp.validate_lqp(txn)
    return txn

def _effect_bindings(effect: Union[ir.Output, ir.Update]) -> list[ir.Value]:
    if isinstance(effect, ir.Output):
        # Unions may not return anything. The generated IR contains a None value when this
        # happens. We ignore it here.

        # TODO: we dont yet handle aliases, so we ignore v[0]
        return [v[1] for v in effect.aliases if v[1] is not None]
    else:
        return list(effect.args)

def _translate_to_decls(ctx: TranslationCtx, rule: ir.Logical) -> list[lqp.Declaration]:
    effects = collect_by_type((ir.Output, ir.Update), rule)
    aggregates = collect_by_type(ir.Aggregate, rule)
    ranks = collect_by_type(ir.Rank, rule)

    # TODO: should this ever actually come in as input?
    if len(effects) == 0:
        return []

    assert len(ranks) == 0 or len(aggregates) == 0, "rules cannot have both aggregates and ranks"

    conjuncts = []
    for task in rule.body:
        if isinstance(task, (ir.Output, ir.Update)):
            continue
        conjuncts.append(_translate_to_formula(ctx, task))

    # Aggregates reduce over the body
    if aggregates or ranks:
        aggr_body = mk_and(conjuncts)
        conjuncts = []
        for aggr in aggregates:
            conjuncts.append(_translate_aggregate(ctx, aggr, aggr_body))
        for rank in ranks:
            conjuncts.append(_translate_rank(ctx, rank, aggr_body))

    return [_translate_effect(ctx, effect, mk_and(conjuncts)) for effect in effects]

def _translate_effect(ctx: TranslationCtx, effect: Union[ir.Output, ir.Update], body: lqp.Formula) -> lqp.Declaration:
    # Handle the bindings
    bindings = _effect_bindings(effect)
    projection, eqs = translate_bindings(ctx, bindings)
    eqs.append(body)
    new_body = mk_and(eqs)

    is_output = isinstance(effect, ir.Output)
    def_name = "output" if is_output else effect.relation.name
    meta_id = effect.id if is_output else effect.relation.id
    rel_id = get_relation_id(ctx, def_name, meta_id)

    # Context bookkeeping
    if is_output:
        ctx.output_ids.append(rel_id)

    return lqp.Def(
        name = rel_id,
        body = mk_abstraction(projection, new_body),
        attrs = [],
        meta = None,
    )

def _translate_rank(ctx: TranslationCtx, rank: ir.Rank, body: lqp.Formula) -> lqp.Formula:
    # Ascending rank is constructed using rel_primitive_sort.
    # Descending rank is constructed from ascending rank by subtracting the rank from the
    # count + 1.
    # Limits are the sort plus a filter on rank <= limit.

    if all(o for o in rank.arg_is_ascending):
        ascending = True
    elif all(not o for o in rank.arg_is_ascending):
        ascending = False
    else:
        raise Exception("Mixed orderings in rank are not supported yet.")

    input_args, input_eqs = translate_bindings(ctx, list(rank.args))

    # Filter out the group-by variables, since they are introduced outside the rank.
    # Input terms are added later below.
    introduced_meta_projs = [arg for arg in rank.projection if arg not in rank.group and arg not in rank.args]
    projected_args, projected_eqs = translate_bindings(ctx, list(introduced_meta_projs))
    body = mk_and([body] + input_eqs + projected_eqs)

    abstr_args = input_args + projected_args
    result_var, result_type = _translate_term(ctx, rank.result)
    assert isinstance(result_var, lqp.Var)
    ranking_terms = [result_var] + [v[0] for v in abstr_args]

    # Rename abstracted args in the body to new variable names
    var_map = {var.name: gen_unique_var(ctx, var.name) for (var, _) in abstr_args}
    body = rename_vars_formula(body, var_map)
    new_abstr_args = [(var_map[var.name], typ) for (var, typ) in abstr_args]

    if ascending and rank.limit == 0:
        return lqp.FFI(
            meta=None,
            name="rel_primitive_sort",
            args=[mk_abstraction(new_abstr_args, body)],
            terms=ranking_terms,
        )
    elif ascending and rank.limit != 0:
        limit_var, limit_type, limit_eq = binding_to_lqp_var(ctx, rank.limit)
        assert limit_eq is not None, "limit_eq should not be None"
        limit_abstr = mk_abstraction([(limit_var, limit_type)], limit_eq)
        top_abstr = mk_abstraction(new_abstr_args, body)

        return lqp.FFI(
            meta=None,
            name="rel_primitive_top",
            args=[top_abstr, limit_abstr],
            terms=ranking_terms,
        )

    # Construct a conjunction of the ranking, a counter for the body, a subtraction
    # of the rank from the count and an addition of 1. Wrap this in an abstraction.
    count_res = gen_unique_var(ctx, "count_res")

    # Add one to the count to account for the rank starting at 1.
    add_one, _, add_one_eq = binding_to_lqp_var(ctx, 1)
    assert add_one_eq is not None, "add_one_eq should not be None"
    one_bigger = gen_unique_var(ctx, "one_bigger")
    addition = lqp.Primitive(
        name="rel_primitive_add",
        terms=[count_res, add_one, one_bigger],
        meta=None
    )

    # Subtract the rank from the count + 1
    asc_rank = gen_unique_var(ctx, "asc_rank")
    subtraction = lqp.Primitive(
        name="rel_primitive_subtract",
        terms=[one_bigger, result_var, asc_rank],
        meta=None
    )

    # Construct the ranking
    desc_ranking_terms = [asc_rank] + [v[0] for v in abstr_args]
    ranking = lqp.FFI(
        meta=None,
        name="rel_primitive_sort",
        args=[mk_abstraction(new_abstr_args, body)],
        terms=desc_ranking_terms,
    )

    # Count the number of rows in the body
    count_var, count_type, count_eq = binding_to_lqp_var(ctx, 1)
    assert count_eq is not None, "count_eq should not be None"
    desc_body = mk_and([body, count_eq])
    aggr_abstr_args = new_abstr_args + [(count_var, count_type)]
    count_aggr = lqp.Reduce(
        op=lqp_operator(
            ctx.unique_names,
            "count",
            "count",
            lqp.PrimitiveType.INT
        ),
        body=mk_abstraction(aggr_abstr_args, desc_body),
        terms=[count_res],
        meta=None
    )

    # Bring it all together and do the maths
    ranking = mk_exists(
        vars=[(asc_rank, result_type), (count_res, result_type), (add_one, result_type), (one_bigger, result_type)],
        value=mk_and([ranking, count_aggr, add_one_eq, addition, subtraction])
    )

    if rank.limit != 0:
        # Wrap with a rank <= limit
        limiter = lqp.Primitive(
            name="rel_primitive_lt_eq",
            terms=[result_var, rank.limit],
            meta=None
        )
        ranking = mk_and([ranking, limiter])

    return ranking

def _translate_aggregate(ctx: TranslationCtx, aggr: ir.Aggregate, body: lqp.Formula) -> Union[lqp.Reduce, lqp.Formula]:
    # TODO: handle this properly
    aggr_name = aggr.aggregation.name
    supported_aggrs = ("sum", "count", "avg", "min", "max", "rel_primitive_solverlib_ho_appl")
    assert aggr_name in supported_aggrs, f"only support {supported_aggrs} for now, not {aggr.aggregation.name}"

    meta_output_terms = []
    meta_input_terms = []

    for (field, arg) in zip(aggr.aggregation.fields, aggr.args):
        if field.input:
            meta_input_terms.append(arg)
        else:
            meta_output_terms.append(arg)

    output_vars = [_translate_term(ctx, term)[0] for term in meta_output_terms]

    body_conjs = [body]
    input_args, input_eqs = translate_bindings(ctx, meta_input_terms)

    # Filter out the group-by variables, since they are introduced outside the aggregation.
    # Input terms are added later below.
    introduced_meta_projs = [arg for arg in aggr.projection if arg not in aggr.group and arg not in meta_input_terms]
    projected_args, projected_eqs = translate_bindings(ctx, list(introduced_meta_projs))
    body_conjs.extend(input_eqs)
    body_conjs.extend(projected_eqs)
    abstr_args: list[Tuple[lqp.Var, lqp.RelType]] = projected_args + input_args

    if aggr_name == "count" or aggr_name == "avg":
        # Count sums up "1"
        one_var, typ, eq = binding_to_lqp_var(ctx, 1)
        assert eq is not None
        body_conjs.append(eq)
        abstr_args.append((one_var, typ))

    body = mk_and(body_conjs)

    # Average needs to wrap the reduce in Exists(Conjunction(Reduce, div))
    if aggr_name == "avg":
        assert len(output_vars) == 1, "avg should only have one output variable"
        output_var = output_vars[0]

        # The average will produce two output variables: sum and count.
        sum_result = gen_unique_var(ctx, "sum")
        count_result = gen_unique_var(ctx, "count")

        # Second to last is the variable we're summing over.
        (sum_var, sum_type) = abstr_args[-2]

        result = lqp.Reduce(
            op=lqp_avg_op(ctx.unique_names, aggr.aggregation.name, sum_var.name, sum_type),
            body=mk_abstraction(abstr_args, body),
            terms=[sum_result, count_result],
            meta=None,
        )

        div = lqp.Primitive(name="rel_primitive_divide", terms=[sum_result, count_result, output_var], meta=None)
        conjunction = mk_and([result, div])

        # Finally, we need to wrap everything in an `exists` to project away the sum and
        # count variables and only keep the result of the division.
        result = mk_exists([(sum_result, sum_type), (count_result, lqp.PrimitiveType.INT)], conjunction)

        return result

    # `input_args`` hold the types of the input arguments, but they may have been modified
    # if we're dealing with a count, so we use `abstr_args` to find the type.
    (aggr_arg, aggr_arg_type) = abstr_args[-1]
    # Group-bys do not need to be handled at all, since they are introduced outside already
    reduce = lqp.Reduce(
        op=lqp_operator(ctx.unique_names, aggr.aggregation.name, aggr_arg.name, aggr_arg_type),
        body=mk_abstraction(abstr_args, body),
        terms=output_vars,
        meta=None
    )
    return reduce

def _translate_to_formula(ctx: TranslationCtx, task: ir.Task) -> lqp.Formula:
    if isinstance(task, ir.Logical):
        conjuncts = [_translate_to_formula(ctx, child) for child in task.body]
        return mk_and(conjuncts)
    elif isinstance(task, ir.Lookup):
        return _translate_to_atom(ctx, task)
    elif isinstance(task, ir.Not):
        return lqp.Not(arg=_translate_to_formula(ctx, task.task), meta=None)
    elif isinstance(task, ir.Exists):
        lqp_vars, conjuncts = translate_bindings(ctx, list(task.vars))
        conjuncts.append(_translate_to_formula(ctx, task.task))
        return mk_exists(lqp_vars, mk_and(conjuncts))
    elif isinstance(task, ir.Construct):
        assert len(task.values) >= 1, "Construct should have at least one value"
        assert isinstance(task.values[0], ir.ScalarType), "Construct should start with a named ScalarType"
        name = task.values[0].name
        terms = [_translate_term(ctx, name)]
        terms.extend([_translate_term(ctx, arg) for arg in task.values[1:]])
        terms.append(_translate_term(ctx, task.id_var))

        return lqp.Primitive(
            name="rel_primitive_hash_tuple_uint128",
            terms=[v for v, _ in terms],
            meta=None
        )
    elif isinstance(task, ir.Union):
        # TODO: handle hoisted vars if needed
        disjs = [_translate_to_formula(ctx, child) for child in task.tasks]
        return mk_or(disjs)
    elif isinstance(task, (ir.Aggregate, ir.Output, ir.Update)):
        # Nothing to do here, handled in _translate_to_decls
        return mk_and([])
    elif isinstance(task, ir.Rank):
        # Nothing to do here, handled in _translate_to_decls
        return mk_and([])
    else:
        raise NotImplementedError(f"Unknown task type (formula): {type(task)}")

# Only used for translating terms on atoms, which can be specialized values.
def _translate_relterm(ctx: TranslationCtx, term: ir.Value) -> Tuple[lqp.RelTerm, lqp.RelType]:
    if isinstance(term, ir.Literal) and term.type == types.Symbol:
        if isinstance(term.value, str):
            return lqp.Specialized(value=term.value, meta=None), type_from_constant(term.value)
        elif isinstance(term.value, int):
            return lqp.Specialized(value=term.value, meta=None), type_from_constant(term.value)
        else:
            raise NotImplementedError(f"Cannot specialize literal of type {type(term.value)}")
    return _translate_term(ctx, term)

def _translate_term(ctx: TranslationCtx, term: ir.Value) -> Tuple[lqp.Term, lqp.RelType]:
    if isinstance(term, ir.Var):
        name = ctx.unique_names.get_name_by_id(term.id, term.name)
        t = meta_type_to_lqp(term.type)
        return mk_var(name), t
    elif isinstance(term, ir.Literal):
        assert isinstance(term.value, lqp.PrimitiveValue), f"expected primitive value, got {type(term.value)}: {term.value}"
        return term.value, type_from_constant(term.value)
    else:
        assert isinstance(term, lqp.PrimitiveValue), \
            f"Cannot translate value {term!r} of type {type(term)} to LQP Term; not a PrimitiveValue."
        return term, type_from_constant(term)

# In the metamodel, type conversions are represented as special relations, whereas in LQP we
# have a dedicated `Cast` node. Eventually we might want to unify these, but for now we use
# this mapping here.
rel_to_cast = {
    "decimal64": lqp.RelValueType.DECIMAL64,
    "decimal128": lqp.RelValueType.DECIMAL128,

    "int_to_float": lqp.PrimitiveType.FLOAT,
    "int_to_decimal64": lqp.RelValueType.DECIMAL64,
    "int_to_decimal128": lqp.RelValueType.DECIMAL128,

    "float_to_int": lqp.PrimitiveType.INT,
    "float_to_decimal64": lqp.RelValueType.DECIMAL64,
    "float_to_decimal128": lqp.RelValueType.DECIMAL128,

    "decimal64_to_float": lqp.PrimitiveType.FLOAT,
    "decimal64_to_int": lqp.PrimitiveType.INT,
    "decimal64_to_decimal128": lqp.RelValueType.DECIMAL128,
    "decimal128_to_float": lqp.PrimitiveType.FLOAT,
    "decimal128_to_int": lqp.PrimitiveType.INT,
    "decimal128_to_decimal64": lqp.RelValueType.DECIMAL64
}

def _translate_to_atom(ctx: TranslationCtx, task: ir.Lookup) -> lqp.Formula:
    # TODO: want signature not name
    rel_name = task.relation.name
    terms = []
    for arg in task.args:
        # Handle varargs, which come wrapped in a tuple.
        if isinstance(arg, tuple):
            for vararg in arg:
                term, _t = _translate_relterm(ctx, vararg)
                terms.append(term)
        else:
            term, _t = _translate_relterm(ctx, arg)
            terms.append(term)

    if rel_builtins.is_builtin(task.relation):
        if task.relation.name in rel_to_cast:
            assert len(terms) == 2, f"expected two terms for cast {task.relation.name}, got {terms}"
            return lqp.Cast(type=rel_to_cast[task.relation.name], input=terms[0], result=terms[1], meta=None)
        elif task.relation.name == "construct_datetime" and len(terms) == 7:
            # construct_datetime does not provide a timezone or milliseconds so we
            # default to 0 milliseconds and UTC timezone.
            lqp_name = relname_to_lqp_name(task.relation.name)
            extended_terms = [*terms[:-1], 0, "UTC", terms[-1]]
            return lqp.Primitive(name=lqp_name, terms=extended_terms, meta=None)
        else:
            lqp_name = relname_to_lqp_name(task.relation.name)
            return lqp.Primitive(name=lqp_name, terms=terms, meta=None)

    if helpers.is_external(task.relation):
        return lqp.RelAtom(name=task.relation.name, terms=terms, meta=None)

    rid = get_relation_id(ctx, rel_name, task.relation.id)
    return lqp.Atom(name=rid, terms=terms, meta=None)

def get_relation_id(ctx: TranslationCtx, orig_name: str, metamodel_id: int) -> lqp.RelationId:
    mid_str = str(metamodel_id)
    relation_id = lqp.RelationId(id=lqp_hash(mid_str), meta=None)
    ctx.id_to_orig_name[relation_id] = orig_name
    return relation_id

def translate_bindings(ctx: TranslationCtx, bindings: list[ir.Value]) -> Tuple[list[Tuple[lqp.Var, lqp.RelType]], list[lqp.Formula]]:
    lqp_vars = []
    conjuncts = []
    for binding in bindings:
        lqp_var, typ, eq = binding_to_lqp_var(ctx, binding)
        lqp_vars.append((lqp_var, typ))
        if eq is not None:
            conjuncts.append(eq)

    return lqp_vars, conjuncts

def binding_to_lqp_var(ctx: TranslationCtx, binding: ir.Value) -> Tuple[lqp.Var, lqp.RelType, Union[None, lqp.Formula]]:
    if isinstance(binding, ir.Var):
        var, typ = _translate_term(ctx, binding)
        assert isinstance(var, lqp.Var)
        return var, typ, None
    else:
        # Constant in this case
        assert isinstance(binding, (lqp.PrimitiveValue, ir.Literal)), f"expected primitive value, got {type(binding)}: {binding}"
        value, typ = _translate_term(ctx, binding)

        var = gen_unique_var(ctx, "cvar")
        eq = lqp.Primitive(name="rel_primitive_eq", terms=[var, value], meta=None)
        return var, typ, eq



import json
from pathlib import Path
from datetime import datetime, timezone
from typing import Iterable, Union


# ─── Module-level in-memory cache and version ──────────────────────────────
_IN_MEMORY_CACHE: dict[str, dict] = {}
METADATA_VERSION = "0.1.0"  # bump when schema changes

class GraphIndexCache:
    def __init__(
        self,
        user: str,
        model: str,
        freshness_mins: Union[int, None],
        sources: Iterable[str],
        metadata_path: str = "metadata.json",
    ):
        self.metadata_path = Path(metadata_path)
        self.key = f"{user}:{model}"
        self.freshness_mins = freshness_mins
        self.sources = list(sources)
        self.using_cache = False

        # Load metadata into memory: prefer in-memory cache, then file
        raw = _IN_MEMORY_CACHE.get(self.key)
        if raw is not None:
            # Use the notebook/session in-memory entry
            self._metadata = {
                "version": METADATA_VERSION,
                "cachedIndices": {self.key: raw},
            }
        else:
            # Fall back to reading from disk
            try:
                with open(self.metadata_path, "r", encoding="utf-8") as f:
                    md = json.load(f)
            except (FileNotFoundError, json.JSONDecodeError):
                md = {}
            # Discard if version mismatched
            if md.get("version") != METADATA_VERSION:
                md = {"version": METADATA_VERSION, "cachedIndices": {}}
            self._metadata = md

    def is_valid(self) -> bool:
        """
        Return True if we have a valid, unexpired cache for this user:model
        covering at least the given sources.
        """
        entry = self._metadata.get("cachedIndices", {}).get(self.key)
        if not entry or self.freshness_mins is None or self.freshness_mins <= 0:
            return False

        # timestamp check
        ts = entry.get("last_use_index_update_on", "")
        try:
            last_ts = datetime.fromisoformat(ts.replace("Z", "+00:00"))
        except Exception:
            return False

        if (
            datetime.now(timezone.utc) - last_ts
        ).total_seconds() > self.freshness_mins * 60:
            return False

        # sources subset check
        if not set(self.sources).issubset(set(entry.get("Sources", []))):
            return False

        self.using_cache = True
        return True

    def choose_sources(self) -> list[str]:
        """
        Return [] if cache is valid, else the full list of sources.
        """
        return [] if self.is_valid() else list(self.sources)

    def record_update(self, sources_info: dict) -> None:
        """
        Write a fresh metadata entry when the index is rebuilt.
        Always update the in-memory cache, then also write to file.
        """
        if self.using_cache:
            return

        # Prepare the new entry
        now = datetime.now(timezone.utc).strftime("%Y-%m-%dT%H:%M:%SZ")
        entry = {"last_use_index_update_on": now, "Sources": list(sources_info.keys())}

        # 1) update module-level in-memory cache
        _IN_MEMORY_CACHE[self.key] = entry
        # 2) merge into our in-memory metadata
        self._metadata.setdefault("cachedIndices", {})[self.key] = entry
        # 3) bump version
        self._metadata["version"] = METADATA_VERSION

        # 4) Persist to disk so the cache survives across processes
        parent = self.metadata_path.parent
        if parent and not parent.exists():
            try:
                parent.mkdir(parents=True, exist_ok=True)
            except OSError:
                return  # cannot create directory → skip file write

        # 5) write out
        try:
            with open(self.metadata_path, "w", encoding="utf-8") as f:
                json.dump(self._metadata, f, indent=2)
        except OSError:
            pass



from __future__ import annotations
import json
import random
import time
from typing import TypeVar
import urllib.parse
import re
from datetime import datetime, timezone

import requests

from relationalai.errors import RAIException

# replace the values of the URL parameters that start with X-Amz- with XXX
def scrub_url(url):
    parsed = urllib.parse.urlparse(url)
    parsed_qs = urllib.parse.parse_qs(parsed.query)
    for key in parsed_qs:
        if key.startswith("X-Amz-"):
            parsed_qs[key] = ["XXX"]
    new_qs = urllib.parse.urlencode(parsed_qs, doseq=True)
    return urllib.parse.urlunparse(parsed._replace(query=new_qs))

def find_urls(string):
    url_pattern = re.compile(r'http[s]?://(?:[a-zA-Z]|[0-9]|[$-_@.&+]|[!*\\(\\),]|(?:%[0-9a-fA-F][0-9a-fA-F]))+')
    urls = re.findall(url_pattern, string)
    return urls

def scrub_urls(string, urls):
    for url in urls:
        # replace with scrubbed version
        string = string.replace(url, scrub_url(url))
    return string

E = TypeVar("E", bound=BaseException)

def scrub_exception(exception: E) -> E|RAIException:
    exception_str = str(exception)
    urls = find_urls(exception_str)
    if urls:
        return RAIException(scrub_urls(exception_str, urls))
    return exception

def wrap_with_request_id(error: requests.RequestException) -> RAIException:
    original_message = str(error)
    try:
        if error.response is not None:
            request_id = error.response.headers['x-amz-request-id']
            return RAIException(f"{original_message} s3 request id: {request_id}")
        return RAIException(original_message)
    except Exception:
        return RAIException(original_message)

def escape_for_f_string(code: str) -> str:
    return (
        code
        .replace("\\", "\\\\")
        .replace("{", "{{")
        .replace("}", "}}")
        .replace("\n", "\\n")
        .replace('"', '\\"')
        .replace("'", "\\'")
    )

def escape_for_sproc(code: str) -> str:
    return code.replace("$$", "\\$\\$")

# @NOTE: `overhead_rate` should fall between 0.05 and 0.5 depending on how time sensitive / expensive the operation in question is.
def poll_with_specified_overhead(
    f,
    overhead_rate: float, # This is the percentage of the time we've already waited before we'll poll again.
    start_time: float | None = None,
    timeout: int | None = None,
    max_tries: int | None = None,
    max_delay: float = 120,
    min_delay: float = 0.1
):
    if overhead_rate < 0:
        raise ValueError("overhead_rate must be non-negative")

    if start_time is None:
        start_time = time.time()

    tries = 0
    max_time = time.time() + timeout if timeout else None

    while True:
        if f():
            break

        current_time = time.time()

        if max_tries is not None and tries >= max_tries:
            raise Exception(f'max tries {max_tries} exhausted')

        if max_time is not None and current_time >= max_time:
            raise Exception(f'timed out after {timeout} seconds')

        duration = (current_time - start_time) * overhead_rate
        duration = max(min(duration, max_delay), min_delay, 0)

        time.sleep(duration)
        tries += 1

def get_with_retries(
    session: requests.Session,
    url: str,
    max_retries: int = 3,
    backoff_factor = 2,
    min_backoff_s = 2,
):
    attempt = 1
    jitter_s = random.uniform(0.5, 1.5)
    delay_s = min_backoff_s
    while True:
        res = session.get(url)
        attempt += 1
        if 500 <= res.status_code <= 504:
            if attempt > max_retries:
                res.raise_for_status()
            else:
                delay_s *= backoff_factor
                time.sleep(delay_s + jitter_s)
                continue
        return res

class ParseError(Exception):
    def __init__(self, identifier):
        super().__init__(f"\nCould not parse fully-qualified name; make sure the identifier is fully qualified and valid:\n{identifier}\n\n")

class IdentityParser:
    """
    Parse a fully-qualified name into its parts and normalize it.
    """
    SF_ID_REGEX = re.compile(r'^[A-Za-z_][A-Za-z0-9_$]*$')

    def __init__(self, identifier: str, require_all_parts: bool = False, force_upper_case: bool = True):
        """
        Args:
            identifier (str): The fully-qualified name to parse.
            require_all_parts (bool): If True, raise an error if any part of the identifier is missing.
            force_upper_case (bool): If True, force all non unique (not double quoted) parts of the identifier to be upper case as per SF standard.
        """
        self.identifier = identifier

        self.db_part = None
        self.schema_part = None
        self.entity_part = None
        self.identity = None
        self._is_complete = False
        self.has_double_quoted_identifier = False
        self.require_all_parts = require_all_parts
        self.force_upper_case = force_upper_case

        self._parse()

    @property
    def db(self):
        return self.db_part['part'] if self.db_part else None

    @property
    def schema(self):
        return self.schema_part['part'] if self.schema_part else None

    @property
    def entity(self):
        return self.entity_part['part'] if self.entity_part else None

    @property
    def is_complete(self):
        return self._is_complete

    def _parse(self):
        self.db_part = self._get_part(self.identifier)
        if self.db_part['next']:
            self.schema_part = self._get_part(self.db_part['next'])
            if self.schema_part['next']:
                self.entity_part = self._get_part(self.schema_part['next'])
                self._is_complete = True

        if self.require_all_parts and (not self.db or not self.schema or not self.entity):
            raise ParseError(self.identifier)

        normalized_parts = [self._normalize_id(x) for x in [self.db_part, self.schema_part, self.entity_part] if x]
        self.identity = '.'.join(part for part in normalized_parts if part is not None)

        # Set has_double_quoted_identifier to True if self.identity contains double quotes
        self.has_double_quoted_identifier = '"' in self.identity if self.identity else False

    def _get_part(self, string: str) -> dict:
        if len(string) == 0:
            return {'part': None, 'next': None}

        if string[0] == '"':
            return self._get_quoted_part(string)
        else:
            return self._get_normal_part(string)

    def _get_normal_part(self, string: str) -> dict:
        ret = ""
        i = 0
        for char in string:
            i += 1
            if char == '.':
                break
            ret += char

        if self.SF_ID_REGEX.match(ret) and self.force_upper_case:
            ret = ret.upper()
        return {'part': ret, 'quoted': False, 'next': string[i:]}

    def _get_quoted_part(self, string: str):
        ret = ""
        i = 1
        while i < len(string):
            char = string[i]
            next_char = string[i + 1] if i + 1 < len(string) else None
            if char == '"':
                if next_char == '.':
                    i += 2
                    break
                elif next_char == '"':
                    ret += '"'
                    i += 1
                elif next_char:
                    raise ParseError(string)
            else:
                ret += char
            i += 1
        return {'part': ret, 'quoted': True, 'next': string[i:]}

    def _normalize_id(self, dict: dict):
        if dict:
            if self.SF_ID_REGEX.match(dict['part']) and not dict['quoted']:
                return dict['part']
            else:
                return '"' + dict['part'].replace('"', '""') + '"'
        return None

    def to_list(self) -> list[str]:
        """
        Return a list of all the fully-qualified name parts including the fqn.
        """
        return [part['part'] for part in [self.db_part, self.schema_part, self.entity_part, {'part': self.identity}] if part is not None]

    @staticmethod
    def to_sql_value(str: str):
        """
        Remove quotes, escape single quotes, and add single quotes around the string.
        """
        _str = str
        if _str.startswith('"') and _str.endswith('"'):
            _str = _str[1:-1]
        return "'" + _str.replace("'", "''") + "'"

def safe_json_loads(s: str):
    return json.loads(s) if s.strip() else {}

def sanitize_module_name(raw_name: str) -> str:
    """
    Turn any Snowflake-normalized name into a valid Python module identifier:
      - lowercase
      - non-alphanumerics => underscore
      - no leading digit
      - collapse multiple underscores
      - strip leading/trailing underscores
      - fallback to 'module' if empty
    """
    name = raw_name.lower()
    # replace anything except a–z, 0–9, or _ with _
    name = re.sub(r"[^0-9a-z_]", "_", name)
    # ensure it doesn’t start with a digit
    if re.match(r"^\d", name):
        name = f"_{name}"
    return name

def ms_to_timestamp(ms):
    if not ms:
        return None
    dt = datetime.fromtimestamp(int(ms) / 1000, tz=timezone.utc).astimezone()  # convert to local time
    return dt.strftime('%Y-%m-%d %H:%M:%S.%f')[:-3] + dt.strftime(' %z')



from relationalai.early_access.metamodel.compiler import Pass
from relationalai.early_access.metamodel import ir, builtins as rel_builtins, factory as f, visitor
from relationalai.early_access.metamodel.typer import Checker, InferTypes
from relationalai.early_access.metamodel import helpers, types
from relationalai.early_access.metamodel.rewrite import Splinter
from relationalai.early_access.metamodel.util import FrozenOrderedSet

from relationalai.early_access.metamodel.rewrite import Flatten
# TODO: Move this into metamodel.rewrite
from relationalai.early_access.rel.rewrite import QuantifyVars, CDC

from relationalai.early_access.lqp.utils import is_constant

import datetime
from decimal import Decimal as PyDecimal
from typing import cast, List, Sequence, Tuple, Union

def lqp_passes() -> list[Pass]:
    return [
        Checker(),
        InferTypes(),
        CDC(),
        # Broken
        # ExtractCommon(),
        Flatten(),
        QuantifyVars(),
        Splinter(), # Adds missing existentials + splits multi-headed rules into single rules
        UnifyDefinitions(),
        EliminateData(),  # Turns Data nodes into ordinary relations.
        EliminateValueTypeConstants(),
        DeduplicateVars(),  # Deduplicates vars in Updates and Outputs.
    ]

class UnifyDefinitions(Pass):
    def rewrite(self, model: ir.Model, options:dict={}) -> ir.Model:
        multidef_relations = self.get_multidef_relations(model)
        for relation in multidef_relations:
            model = self.rename_multidef(model, relation)
        return model

    def get_multidef_relations(self, model: ir.Model) -> set[ir.Relation]:
        seen = set()
        result = set()
        root = cast(ir.Logical, model.root)
        for task in root.body:
            task = cast(ir.Logical, task)
            for subtask in task.body:
                if isinstance(subtask, ir.Update):
                    assert subtask.effect == ir.Effect.derive, "only derive updates supported yet"
                    name = subtask.relation
                    if name.id in seen:
                        result.add(name)
                    seen.add(name.id)
        return result

    def rename_multidef(self, model: ir.Model, relation: ir.Relation) -> ir.Model:
        root = cast(ir.Logical, model.root)

        new_subtasks = []
        new_relations = []
        generated_relation_names = {}
        total_ct = 0

        # Rename occurrences of the relation in the model
        for subtask in root.body:
            subtask = cast(ir.Logical, subtask)
            new_subsubtasks: list[ir.Task] = []
            for subsubtask in subtask.body:
                if isinstance(subsubtask, ir.Update):
                    assert subsubtask.effect == ir.Effect.derive, "only derive updates supported yet"
                    name = subsubtask.relation
                    if name.id == relation.id:
                        total_ct += 1
                        # TODO: this needs to be unique btw (gensym)
                        new_name = f"{relation.name}_{total_ct}"

                        # Check if we already generated this relation name. If we did, just
                        # reuse it, otherwise we end up with undefined relation IDs.
                        if new_name in generated_relation_names:
                            new_relation = generated_relation_names[new_name]
                        else:
                            new_relation = ir.Relation(
                                new_name,
                                name.fields,
                                name.requires,
                            )
                            new_relations.append(new_relation)
                            generated_relation_names[new_name] = new_relation

                        new_subsubtask = ir.Update(
                            subsubtask.engine,
                            new_relation,
                            subsubtask.args,
                            subsubtask.effect,
                        )
                        new_subsubtasks.append(new_subsubtask)
                    else:
                        new_subsubtasks.append(subsubtask)
                else:
                    new_subsubtasks.append(subsubtask)

            new_subtask = ir.Logical(
                subtask.engine,
                subtask.hoisted,
                tuple(new_subsubtasks),
            )
            new_subtasks.append(new_subtask)

        assert total_ct > 0, f"should have found at least one definition for {relation.name}"

        args = []
        for field in relation.fields:
            args.append(ir.Var(field.type, field.name))

        # Also add the new definition, using the existing relation
        new_update = ir.Update(
            root.engine,
            relation,
            tuple(args),
            ir.Effect.derive,
        )

        logical_tasks = []
        lookups = []
        for new_relation in new_relations:
            new_lookup = ir.Lookup(
                root.engine,
                new_relation,
                tuple(args),
            )
            lookups.append(new_lookup)

        disj = ir.Union(
            root.engine,
            tuple(),
            tuple(lookups),
        )
        logical_tasks.append(disj)
        logical_tasks.append(new_update)
        new_logical = ir.Logical(
            root.engine,
            root.hoisted,
            tuple(logical_tasks),
        )
        new_subtasks.append(new_logical)

        new_root = ir.Logical(root.engine, root.hoisted, tuple(new_subtasks))
        model = ir.Model(
            model.engines,
            model.relations | new_relations,
            model.types,
            new_root,
        )
        return model

# We don't have a proto representation for these, so we rewrite them to be constructed directly
class EliminateValueTypeConstants(Pass):
    def rewrite(self, model: ir.Model, options:dict={}) -> ir.Model:
        p = RewriteBadLiterals()
        return p.walk(model)

class RewriteBadLiterals(visitor.Rewriter):
    def rewrite_bad_literals(self, args):
        vars_to_existify = []
        new_conjs = []
        new_args = []

        for arg in args:
            if is_constant(arg, datetime.datetime):
                if isinstance(arg, ir.Literal):
                    arg = arg.value
                new_var = f.var("dt_var", types.DateTime)
                new_args.append(new_var)
                vars_to_existify.append(new_var)

                year = arg.year
                month = arg.month
                day = arg.day
                hour = arg.hour
                minute = arg.minute
                second = arg.second

                lookup = f.lookup(
                        rel_builtins.construct_datetime,
                        tuple([
                            f.literal(year),
                            f.literal(month),
                            f.literal(day),
                            f.literal(hour),
                            f.literal(minute),
                            f.literal(second),
                            new_var,
                        ])
                    )
                new_conjs.append(lookup)
            elif is_constant(arg, datetime.date):
                if isinstance(arg, ir.Literal):
                    arg = arg.value
                new_var = f.var("dt_var", types.Date)
                new_args.append(new_var)
                vars_to_existify.append(new_var)

                year = arg.year
                month = arg.month
                day = arg.day

                lookup = f.lookup(
                        rel_builtins.construct_date,
                        tuple([
                            f.literal(year),
                            f.literal(month),
                            f.literal(day),
                            new_var,
                        ])
                    )
                new_conjs.append(lookup)
            elif is_constant(arg, PyDecimal):
                if isinstance(arg, ir.Literal):
                    arg = arg.value
                new_var = f.var("dec_var", types.Decimal128)
                new_args.append(new_var)
                vars_to_existify.append(new_var)

                lookup = f.lookup(
                    rel_builtins.parse_decimal,
                    tuple([
                        f.literal(128, types.Symbol),
                        f.literal(10, types.Symbol),
                        f.literal(str(arg)),
                        new_var
                    ]),
                )
                new_conjs.append(lookup)
            else:
                new_args.append(arg)

        return new_args, vars_to_existify, new_conjs

    def handle_lookup(self, node: ir.Lookup, parent: ir.Node) -> Union[ir.Lookup, ir.Exists]:
        args = node.args
        new_args, vars_to_existify, new_conjs = self.rewrite_bad_literals(args)

        if len(vars_to_existify) == 0:
            return node

        new_lookup = f.lookup(
            node.relation,
            tuple(new_args),
        )
        new_conjs.append(new_lookup)

        result = f.exists(
            vars_to_existify,
            f.logical(tuple(new_conjs)),
        )

        return result

    def handle_aggregate(self, node: ir.Aggregate, parent: ir.Node) -> Union[ir.Aggregate, ir.Exists]:
        args = node.args
        new_args, vars_to_existify, new_conjs = self.rewrite_bad_literals(args)

        if len(vars_to_existify) == 0:
            return node

        new_aggregate = node.reconstruct(args=tuple(new_args))
        new_conjs.append(new_aggregate)

        result = f.exists(
            vars_to_existify,
            f.logical(tuple(new_conjs)),
        )

        return result

    def handle_output(self, node: ir.Output, parent: ir.Node) -> Union[ir.Output, ir.Logical]:
        args = [v[1] for v in node.aliases]
        new_args, vars_to_existify, new_conjs = self.rewrite_bad_literals(args)

        if len(vars_to_existify) == 0:
            return node

        new_aliases = []
        for i, alias in enumerate(node.aliases):
            new_aliases.append((alias[0], new_args[i]))

        new_output = node.reconstruct(aliases=tuple(new_aliases))
        new_conjs.append(new_output)

        return f.logical(new_conjs)

    def handle_update(self, node: ir.Update, parent: ir.Node) -> Union[ir.Update, ir.Logical]:
        args = node.args
        new_args, vars_to_existify, new_conjs = self.rewrite_bad_literals(args)

        if len(vars_to_existify) == 0:
            return node

        new_update = node.reconstruct(args=tuple(new_args))
        new_conjs.append(new_update)
        return f.logical(new_conjs)

    def handle_construct(self, node: ir.Construct, parent: ir.Node) -> Union[ir.Construct, ir.Exists]:
        values = node.values
        new_values, vars_to_existify, new_conjs = self.rewrite_bad_literals(values)

        if len(vars_to_existify) == 0:
            return node

        new_construct = node.reconstruct(values=tuple(new_values))
        new_conjs.append(new_construct)
        return f.exists(
            vars_to_existify,
            f.logical(tuple(new_conjs)),
        )


# Creates intermediary relations for all Data nodes and replaces said Data nodes
# with a Lookup into these created relations.
class EliminateData(Pass):
    def rewrite(self, model: ir.Model, options:dict={}) -> ir.Model:
        r = self.DataRewriter()
        return r.walk(model)

    # Does the actual work.
    class DataRewriter(visitor.Rewriter):
        new_relations: list[ir.Relation]
        new_updates: list[ir.Logical]
        # Counter for naming new relations.
        # It must be that new_count == len new_updates == len new_relations.
        new_count: int

        def __init__(self):
            self.new_relations = []
            self.new_updates = []
            self.new_count = 0
            super().__init__()

        # Create a new intermediary relation representing the Data (and pop it in
        # new_updates/new_relations) and replace this Data with a Lookup of said
        # intermediary.
        def handle_data(self, node: ir.Data, parent: ir.Node) -> ir.Lookup:
            self.new_count += 1
            intermediary_name = f"formerly_Data_{self.new_count}"

            intermediary_relation = f.relation(
                intermediary_name,
                [f.field(v.name, v.type) for v in node.vars]
            )
            self.new_relations.append(intermediary_relation)

            intermediary_update = f.logical([
                # For each row (union), equate values and their variable (logical).
                f.union(
                    [
                        f.logical(
                            [
                                f.lookup(rel_builtins.eq, [val, var])
                                for (val, var) in zip(row, node.vars)
                            ],
                            hoisted = node.vars,
                        )
                        for row in node
                    ],
                    hoisted = node.vars,
                ),
                # And pop it back into the relation.
                f.update(intermediary_relation, node.vars, ir.Effect.derive),
            ])
            self.new_updates.append(intermediary_update)

            replacement_lookup = f.lookup(intermediary_relation, node.vars)

            return replacement_lookup

        # Walks the model for the handle_data work then updates the model with
        # the new state.
        def handle_model(self, model: ir.Model, parent: None):
            walked_model = super().handle_model(model, parent)
            assert len(self.new_relations) == len(self.new_updates) and self.new_count == len(self.new_relations)

            # This is okay because its LQP.
            assert isinstance(walked_model.root, ir.Logical)
            root_logical = cast(ir.Logical, walked_model.root)

            # We may need to add the new intermediaries from handle_data to the model.
            if self.new_count  == 0:
                return model
            else:
                return ir.Model(
                    walked_model.engines,
                    walked_model.relations | self.new_relations,
                    walked_model.types,
                    ir.Logical(
                        root_logical.engine,
                        root_logical.hoisted,
                        root_logical.body + tuple(self.new_updates),
                        root_logical.annotations,
                    ),
                    walked_model.annotations,
                )

# Deduplicate Vars in Updates and Outputs.
class DeduplicateVars(Pass):
    def rewrite(self, model: ir.Model, options:dict={}) -> ir.Model:
        r = self.VarDeduplicator()
        return r.walk(model)

    # Return 1) a new list of Values with no duplicates (at the object level) and
    # 2) equalities between any original Value and a deduplicated Value.
    @staticmethod
    def dedup_values(vals: Sequence[ir.Value]) -> Tuple[List[ir.Value], List[ir.Lookup]]:
        # If a var is seen more than once, it is a duplicate and we will create
        # a new Var and equate it with the seen one.
        seen_vars = set()

        new_vals = []
        eqs = []

        for i, val in enumerate(vals):
            # Duplicates can only occur within Vars.
            # TODO: we don't know for sure if these are the only relevant cases.
            if isinstance(val, ir.Default) or isinstance(val, ir.Var):
                var = val if isinstance(val, ir.Var) else val.var
                if var in seen_vars:
                    new_var = ir.Var(var.type, var.name + "_dup_" + str(i))
                    new_val = new_var if isinstance(val, ir.Var) else ir.Default(new_var, val.value)
                    new_vals.append(new_val)
                    eqs.append(f.lookup(rel_builtins.eq, [new_var, var]))
                else:
                    seen_vars.add(var)
                    new_vals.append(val)
            else:
                # No possibility of problematic duplication.
                new_vals.append(val)

        return new_vals, eqs

    # Returns a reconstructed output with no duplicate variable objects
    # (dedup_values) and now necessary equalities between any two previously
    # duplicate variables.
    @staticmethod
    def dedup_output(output: ir.Output) -> List[Union[ir.Output, ir.Lookup]]:
        vals = helpers.output_values(output.aliases)
        deduped_vals, req_lookups = DeduplicateVars.dedup_values(vals)
        # Need the names so we can recombine.
        alias_names = [alias[0] for alias in output.aliases]
        new_output = output.reconstruct(
            output.engine,
            FrozenOrderedSet([tuple(alias) for alias in zip(alias_names, deduped_vals)]),
            output.keys,
            output.annotations,
        )
        return [new_output] + req_lookups

    # Returns a reconstructed update with no duplicate variable objects
    # (dedup_values) and now necessary equalities between any two previously
    # duplicate variables.
    @staticmethod
    def dedup_update(update: ir.Update) -> List[Union[ir.Update, ir.Lookup]]:
        deduped_vals, req_lookups = DeduplicateVars.dedup_values(update.args)
        new_update = update.reconstruct(
            update.engine,
            update.relation,
            tuple(deduped_vals),
            update.effect,
            update.annotations,
        )
        return [new_update] + req_lookups

    # Does the actual work.
    class VarDeduplicator(visitor.Rewriter):
        def __init__(self):
            super().__init__()

        # We implement handle_logical instead of handle_update/handle_output
        # because in addition to modifying said update/output we require new
        # lookups (equality between original and deduplicated variables).
        def handle_logical(self, node: ir.Logical, parent: ir.Node):
            # In order to recurse over subtasks.
            node = super().handle_logical(node, parent)

            new_body = []
            for subtask in node.body:
                if isinstance(subtask, ir.Output):
                    new_body.extend(DeduplicateVars.dedup_output(subtask))
                elif isinstance(subtask, ir.Update):
                    new_body.extend(DeduplicateVars.dedup_update(subtask))
                else:
                    new_body.append(subtask)

            return node.reconstruct(
                node.engine,
                node.hoisted,
                tuple(new_body),
                node.annotations
            )



"""
Type checking for the IR.
"""
from dataclasses import dataclass, field
from typing import Optional, List, Union as PyUnion, Tuple, cast

from relationalai.early_access.metamodel.util import OrderedSet, ordered_set
from relationalai.early_access.metamodel import ir, types, visitor, compiler
import rich


@dataclass
class CheckError:
    """A model well-formedness error."""
    msg: str
    node: ir.Node

    def __str__(self):
        return f"[red bold][Model Error][/red bold] {self.msg}: [white]{str(self.node).strip()}[/white]"


@dataclass
class CheckEnv:
    """Environment for checking that a model is well-formed."""

    # The model being type-checked.
    model: ir.Model

    # Diagnostics. For now, this is just strings.
    diags: List[CheckError]

    # How verbose to be with debug info, 0 for off.
    verbosity: int

    def __init__(self, model: ir.Model, verbosity: int=0):
        self.model = model
        self.diags = []
        self.verbosity = verbosity

    def _complain(self, node: ir.Node, msg: str):
        """Report an error."""
        self.diags.append(CheckError(msg, node))


@dataclass
class Checker(compiler.Pass):
    """
    A pass that checks that a model is well-formed.
    Diagnostics are reported for ill-formed models.
    """

    # How verbose to be with debug output, 0 is off.
    verbosity: int = field(default=0, init=False)

    def rewrite(self, model: ir.Model, options:dict={}) -> ir.Model:
        # Type checking.
        env = CheckEnv(model, self.verbosity)
        checker = CheckModel(env)
        model.accept(checker)

        # Report any well-formedness errors
        if env.diags:
            for diag in env.diags:
                rich.print(str(diag))
            if self.verbosity:
                rich.print("[dim]-----[/dim]")
                rich.print(str(model))

        return model


@dataclass
class CheckModel(visitor.DAGVisitor):
    """
    A visitor that checks that a model is well-formed.
    """
    def __init__(self, env: CheckEnv):
        super().__init__()
        self.env = env

    def visit_model(self, node: ir.Model, parent: Optional[ir.Node]=None) -> None:
        env = self.env
        model = node

        # Check that all types are present in the model.
        @dataclass
        class CheckPresence(visitor.Visitor):
            model: ir.Model = field(init=True)
            env: CheckEnv = field(init=True)
            def visit_scalartype(self, node: ir.ScalarType, parent: Optional[ir.Node]=None) -> None:
                if node not in self.model.types:
                    self.env.diags.append(CheckError(f"Type {ir.node_to_string(node).strip()} not found in the model.", node))
            def visit_listtype(self, node: ir.ListType, parent: Optional[ir.Node]=None) -> None:
                if node not in self.model.types:
                    self.env.diags.append(CheckError(f"Type {ir.node_to_string(node).strip()} not found in the model.", node))
            def visit_uniontype(self, node: ir.UnionType, parent: Optional[ir.Node]=None) -> None:
                if node not in self.model.types:
                    self.env.diags.append(CheckError(f"Type {ir.node_to_string(node).strip()} not found in the model.", node))
            def visit_tupletype(self, node: ir.TupleType, parent: Optional[ir.Node]=None) -> None:
                if node not in self.model.types:
                    self.env.diags.append(CheckError(f"Type {ir.node_to_string(node).strip()} not found in the model.", node))

        for t in model.types:
            t.accept(CheckPresence(model, env))

        # Check that scalar type names are unique.
        type_names = set()
        for t in model.types:
            if isinstance(t, ir.ScalarType):
                if t.name in type_names:
                    env.diags.append(CheckError(f"Type {t} has is not unique {t.name}.", t))
                type_names.add(t.name)

        # Check that the types are acyclic and that the supertype relation is acyclic.
        def check_type_cycles(t: ir.Type, subtypes: OrderedSet[ir.Type]=ordered_set(), containers: OrderedSet[ir.Type]=ordered_set()) -> None:
            if t in subtypes:
                env.diags.append(CheckError(f"Type {t} is involved in a subtyping cycle with {', '.join(str(s) for s in (subtypes - {t}))}.", t))
            if t in containers:
                env.diags.append(CheckError(f"Type {t} is infinitely recursive. It is contained within itself.", t))
            if isinstance(t, ir.ScalarType):
                for s in t.super_types:
                    check_type_cycles(s, subtypes | {t}, containers)
            if isinstance(t, ir.ListType):
                check_type_cycles(t.element_type, subtypes, containers | {t})
            if isinstance(t, ir.UnionType):
                for s in t.types:
                    check_type_cycles(s, subtypes, containers | {t})
            if isinstance(t, ir.TupleType):
                for s in t.types:
                    check_type_cycles(s, subtypes, containers | {t})

        # Check that the types are well-formed and acyclic.
        for t in model.types:
            check_type_cycles(t)
            # Lists must have scalar element types.
            if isinstance(t, ir.ListType):
                if not isinstance(t.element_type, ir.ScalarType):
                    env.diags.append(CheckError(f"Type {ir.type_to_string(t)} is an list type, but extends non-scalar element type {ir.type_to_string(t.element_type)}.", t))
            # Tuples must have scalar element types.
            elif isinstance(t, ir.TupleType):
                for s in t.types:
                    if not isinstance(s, ir.ScalarType):
                        env.diags.append(CheckError(f"Type {ir.type_to_string(t)} is an tuple type, but extends non-scalar element type {ir.type_to_string(s)}.", t))
            # Union types must have scalar element types.
            elif isinstance(t, ir.UnionType):
                for s in t.types:
                    if not isinstance(s, ir.ScalarType):
                        env.diags.append(CheckError(f"Type {ir.type_to_string(t)} is an union type, but extends non-scalar element type {ir.type_to_string(s)}.", t))
            elif isinstance(t, ir.ScalarType):
                # Supertypes of scalar types must be scalar types.
                for s in t.super_types:
                    if not isinstance(s, ir.ScalarType):
                        env.diags.append(CheckError(f"Type {ir.type_to_string(t)} is an scalar type, but extends a non-scalar type {ir.type_to_string(s)}.", t))
                if types.is_any(t):
                    # The Any type should not have supertypes.
                    if len(t.super_types) > 0:
                        env.diags.append(CheckError(f"Type {ir.type_to_string(t)} is the Any type, but extends some supertypes. The Any type should not have supertypes.", t))
                elif types.is_value_base_type(t):
                    # Value base types should not have supertypes.
                    if len(t.super_types) > 0:
                        if len(t.super_types) != 1 or (t.super_types[0] != types.Number and t.super_types[0] != types.String):
                            env.diags.append(CheckError(f"Type {ir.type_to_string(t)} is an value base type, but extends one or more supertypes. Value base types should not have supertypes (except for Number or String).", t))
                elif types.is_value_type(t):
                    # By construction, a value type should either be a value base type or extend a value base type, so there's no need to check that.
                    # Value types should have exactly one supertype.
                    if len(t.super_types) > 1:
                        env.diags.append(CheckError(f"Type {ir.type_to_string(t)} is an value type, but extends multiple supertypes. User-defined value types must have exactly one supertype.", t))
                    # Value types cannot extend Hash.
                    if types.is_subtype(t, types.Hash):
                        env.diags.append(CheckError(f"Type {ir.type_to_string(t)} is an value type, but extends the entity type `Hash`. Entity types and value types must be disjoint.", t))

        # Recurse to check the rest of the model.
        return super().visit_model(node, parent)

    def visit_engine(self, node: ir.Engine, parent: Optional[ir.Node]=None):
        # Check that each relation requires a subset of the engine capabilities.
        for r in node.relations:
            for c in r.requires:
                if c not in node.capabilities:
                    self.env._complain(node, f"Relation `{r.name}` requires capability {c}, but engine `{node.name}` only has {', '.join([str(x) for x in node.capabilities])}.")
        # Do not recurse. We already checked the relations at the model level.
        pass

    def visit_scalartype(self, node: ir.Type, parent: Optional[ir.Node]=None):
        # Do not recurse.
        pass

    def visit_listtype(self, node: ir.ListType, parent: Optional[ir.Node]=None):
        # List types are allowed before type inference.
        # Type inference will complain if the element is inferred to be a list type.
        # Do not recurse.
        pass

    def visit_relation(self, node: ir.Relation, parent: Optional[ir.Node]=None):
        # don't check relations used in annotations
        if parent and isinstance(parent, ir.Annotation):
            return

        if node not in self.env.model.relations:
            self.env._complain(node, f"Relation `{node.name}` (id={node.id}) is not declared in the model.")
        field_names = [f.name for f in node.fields]
        duplicate_names = [name for name in field_names if field_names.count(name) > 1]
        if duplicate_names:
            self.env._complain(node, f"Relation `{node.name}` has duplicate field names: {', '.join(duplicate_names)}.")
        # Check that all overloads are proper subtypes of the original relation
        for r in node.overloads:
            if r is node:
                self.env._complain(node, f"Relation `{node.name}` has an overload that is itself.")
            if r.name != node.name:
                self.env._complain(r, f"Relation `{node.name}` has an overload that has a different name.")
            if len(r.fields) != len(node.fields):
                self.env._complain(r, f"Relation `{node.name}` has an overload that has a different number of fields.")
            # TODO this test is disabled until we add back Int <: Number
            # from relationalai.early_access.metamodel import helpers
            # if not helpers.relation_is_proper_subtype(r, node):
            #     self.env._complain(r, f"Relation `{r.name}` is an overload but is not a proper subtype of the original relation.")
            if r.overloads:
                self.env._complain(r, f"Relation `{r.name}` has an overload that has an overload.")
            for f1, f2 in zip(r.fields, node.fields):
                if f1.input and not f2.input:
                    self.env._complain(r, f"Relation `{r.name}` has an overload that has an input field that is not present in the original relation.")
                if not f1.input and f2.input:
                    self.env._complain(r, f"Relation `{node.name}` has an input field that is not present in the overload.")

        # Recurse to visit the fields.
        return super().visit_relation(node, parent)

    def visit_var(self, node: ir.Var, parent: Optional[ir.Node]=None):
        # Do not constrain field types to be <: Any.
        if not isinstance(node.type, ir.UnionType) and not isinstance(node.type, ir.ScalarType):
            self.env._complain(node, f"Variable `{node.name}` has type {ir.type_to_string(node.type)}, which is not a scalar or union type.")
        # Do not recurse. No need to visit the type.
        return super().visit_var(node, parent)

    def visit_logical(self, node: ir.Logical, parent: Optional[ir.Node]=None):
        # The hoisted variables should occur in the body.
        for x in node.hoisted:
            if not CheckModel._variable_occurs_in(x, node.body):
                self.env._complain(node, f"Variable {ir.node_to_string(x).strip()} is hoisted but not used in the body of {ir.node_to_string(node).strip()}.")
        return super().visit_logical(node, parent)

    def visit_union(self, node: ir.Union, parent: Optional[ir.Node]=None):
        # The hoisted variables should occur in the body.
        for x in node.hoisted:
            if not CheckModel._variable_occurs_in(x, node.tasks):
                self.env._complain(node, f"Variable {ir.node_to_string(x).strip()} is hoisted but not used in the body of {ir.node_to_string(node).strip()}.")
        return super().visit_union(node, parent)

    def visit_sequence(self, node: ir.Sequence, parent: Optional[ir.Node]=None):
        # The hoisted variables should occur in the body.
        for x in node.hoisted:
            if not CheckModel._variable_occurs_in(x, node.tasks):
                self.env._complain(node, f"Variable {ir.node_to_string(x).strip()} is hoisted but not used in the body of {ir.node_to_string(node).strip()}.")
        return super().visit_sequence(node, parent)

    def visit_match(self, node: ir.Match, parent: Optional[ir.Node]=None):
        # The hoisted variables should occur in the body.
        for x in node.hoisted:
            if not CheckModel._variable_occurs_in(x, node.tasks):
                self.env._complain(node, f"Variable {ir.node_to_string(x).strip()} is hoisted but not used in the body of {ir.node_to_string(node).strip()}.")
        return super().visit_match(node, parent)

    def visit_until(self, node: ir.Until, parent: Optional[ir.Node]=None):
        # The hoisted variables should occur in the body.
        for x in node.hoisted:
            if not CheckModel._variable_occurs_in(x, node.body):
                self.env._complain(node, f"Variable {ir.node_to_string(x).strip()} is hoisted but not used in the body of {ir.node_to_string(node).strip()}.")
        return super().visit_until(node, parent)

    def visit_wait(self, node: ir.Wait, parent: Optional[ir.Node]=None):
        # The hoisted variables should occur in the body.
        for x in node.hoisted:
            if not CheckModel._variable_occurs_in(x, node.check):
                self.env._complain(node, f"Variable {ir.node_to_string(x).strip()} is hoisted but not used in the body of {ir.node_to_string(node).strip()}.")
        return super().visit_wait(node, parent)

    def visit_exists(self, node: ir.Exists, parent: Optional[ir.Node]=None):
        # The quantified variables should occur in the body.
        for x in node.vars:
            if not CheckModel._variable_occurs_in(x, node.task):
                self.env._complain(node, f"Variable {ir.node_to_string(x).strip()} is quantified but not used in the body of {ir.node_to_string(node).strip()}.")
        return super().visit_exists(node, parent)

    def visit_forall(self, node: ir.ForAll, parent: Optional[ir.Node]=None):
        # The quantified variables should occur in the body.
        for x in node.vars:
            if not CheckModel._variable_occurs_in(x, node.task):
                self.env._complain(node, f"Variable {ir.node_to_string(x).strip()} is quantified but not used in the body of {ir.node_to_string(node).strip()}.")
        return super().visit_forall(node, parent)

    def visit_output(self, node: ir.Output, parent: Optional[ir.Node]):
        # Outputs are expected to be children of Logicals.
        if not isinstance(parent, ir.Logical):
            self.env._complain(node, f"Output is not a child of a Logical, but a {type(parent)}.")

        super().visit_output(node, parent)

    @staticmethod
    def _variable_occurs_in(node: ir.VarOrDefault, body: PyUnion[ir.Task, Tuple[ir.Task, ...]]) -> bool:
        if isinstance(body, tuple):
            return any(CheckModel._variable_occurs_in(node, b) for b in body)
        if isinstance(node, ir.Var):
            return node in visitor.collect_by_type(ir.Var, cast(ir.Node, body))
        elif isinstance(node, ir.Default):
            return node.var in visitor.collect_by_type(ir.Var, cast(ir.Node, body))
        else:
            return False

    def visit_loop(self, node: ir.Loop, parent: Optional[ir.Node]=None):
        # The hoisted variables should occur in the body.
        for x in node.hoisted:
            if not CheckModel._variable_occurs_in(x, node.body):
                self.env._complain(node, f"Variable {ir.node_to_string(x).strip()} is hoisted but not used in the body of {ir.node_to_string(node).strip()}.")
        if not CheckModel._variable_occurs_in(node.iter, node.body):
            self.env._complain(node, f"Variable {node.iter} is the loop iterator but is not used in the body of {ir.node_to_string(node).strip()}.")
        return super().visit_loop(node, parent)

    def visit_update(self, node: ir.Update, parent: Optional[ir.Node]=None):
        if len(node.args) != len(node.relation.fields):
            self.env._complain(node, f"{ir.node_to_string(node).strip()} has {len(node.args)} arguments but relation {node.relation.name} has {len(node.relation.fields)} fields")
        # Updates are expected to be children of Logicals.
        if not isinstance(parent, ir.Logical):
            self.env._complain(node, f"Update is not a child of a Logical, but a {type(parent)}.")
        return super().visit_update(node, parent)

    def visit_lookup(self, node: ir.Lookup, parent: Optional[ir.Node]=None):
        if len(node.args) != len(node.relation.fields):
            self.env._complain(node, f"{ir.node_to_string(node).strip()} has {len(node.args)} arguments but relation {node.relation.name} has {len(node.relation.fields)} fields")
        return super().visit_lookup(node, parent)

    def visit_construct(self, node: ir.Construct, parent: Optional[ir.Node]=None):
        if not node.values:
            self.env._complain(node, f"Construct {ir.node_to_string(node).strip()} should have at least 1 value. You cannot construct something from nothing.")
        return super().visit_construct(node, parent)

    def visit_aggregate(self, node: ir.Aggregate, parent: Optional[ir.Node]=None):
        agg = node.aggregation
        if len(agg.fields) < 1:
            self.env._complain(node, f"Aggregation `{agg.name}` should have at least 1 field, found {len(agg.fields)}.")
            return

        inputs = []
        outputs = []
        for f in agg.fields:
            if f.input:
                if outputs:
                    self.env._complain(node, f"Aggregation `{agg.name}` should declare all input fields before any output fields.")
                inputs.append(f)
            else:
                outputs.append(f)

        # Now let's wire up the types.

        # Inputs and outputs.
        if len(node.args) != len(inputs) + len(outputs):
            self.env._complain(node, f"Aggregation `{agg.name}` expects {len(inputs) + len(outputs)} arguments, but has {len(node.args)} arguments instead.")

        return super().visit_aggregate(node, parent)



from __future__ import annotations
from typing import Any
from decimal import Decimal as PyDecimal

from .. import builder as b

def _make_expr(op: str, *args: Any) -> b.Expression:
    return b.Expression(b.Relationship.builtins[op], *args)

# Coerce a number to Decimal64.
def decimal64(value: b.Producer|int|float|PyDecimal) -> b.Expression:
    if isinstance(value, int):
        return _make_expr("int_to_decimal64", value, b.Decimal64.ref("res"))
    if isinstance(value, float):
        return _make_expr("float_to_decimal64", value, b.Decimal64.ref("res"))
    return _make_expr("decimal64", value, b.Decimal64.ref("res"))

# Coerce a number to Decimal128.
def decimal128(value: b.Producer|int|float|PyDecimal) -> b.Expression:
    if isinstance(value, int):
        return _make_expr("int_to_decimal128", value, b.Decimal128.ref("res"))
    if isinstance(value, float):
        return _make_expr("float_to_decimal128", value, b.Decimal128.ref("res"))
    return _make_expr("decimal128", value, b.Decimal128.ref("res"))

def decimal(value: b.Producer|int|float|PyDecimal) -> b.Expression:
    return decimal128(value)

def parse_decimal64(value: b.Producer|str) -> b.Expression:
    return _make_expr("parse_decimal64", value, b.Decimal64.ref("res"))

def parse_decimal128(value: b.Producer|str) -> b.Expression:
    return _make_expr("parse_decimal128", value, b.Decimal128.ref("res"))



from collections import defaultdict
from relationalai.early_access.dsl.adapters.orm.model import ExclusiveInclusiveSubtypeFact, ExclusiveSubtypeFact, \
    ORMRingType
from relationalai.early_access.dsl.adapters.orm.parser import ORMParser
from relationalai.early_access.dsl.core.utils import to_pascal_case
from relationalai.early_access.dsl.orm.constraints import RingType
from relationalai.early_access.dsl.orm.models import Model
from relationalai.early_access.builder import Integer, String, DateTime, Decimal, Date
from relationalai.early_access.dsl.orm.utils import build_relation_name_from_reading


class ORMAdapterQB:
    _datatype_mapping = {
        "AutoCounterNumericDataType": Integer,
        # TODO  must be changed to UnsignedInt when available
        "UnsignedIntegerNumericDataType": Integer,
        "VariableLengthTextDataType": String,
        "DateAndTimeTemporalDataType": DateTime,
        "DecimalNumericDataType": Decimal,
        "DateTemporalDataType": Date
    }

    _ring_type_mapping = {
        ORMRingType.IRREFLEXIVE: RingType.IRREFLEXIVE,
        ORMRingType.ANTISYMMETRIC: RingType.ANTISYMMETRIC,
        ORMRingType.ASYMMETRIC: RingType.ASYMMETRIC,
        ORMRingType.INTRANSITIVE: RingType.INTRANSITIVE,
        ORMRingType.STRONGLY_INTRANSITIVE: RingType.STRONGLY_INTRANSITIVE,
        ORMRingType.ACYCLIC: RingType.ACYCLIC,
        ORMRingType.PURELY_REFLEXIVE: RingType.PURELY_REFLEXIVE,
        ORMRingType.REFLEXIVE: RingType.REFLEXIVE,
        ORMRingType.SYMMETRIC: RingType.SYMMETRIC,
        ORMRingType.TRANSITIVE: RingType.TRANSITIVE
    }

    def __init__(self, orm_file_path: str):
        self._parser = ORMParser(orm_file_path)
        self._relationship_role_value_constraints = defaultdict()
        self._relationships = {}
        self.model = self.orm_to_model()

    def orm_to_model(self):
        model = Model(self._parser.model_name())

        self._add_value_types(model)
        self._add_entity_types(model)
        self._add_subtype_relationships(model)
        self._add_relationships(model)
        self._add_external_identifying_relationships(model)
        self._add_role_value_constraints(model)
        self._add_role_constraints(model)
        self._add_ring_constraints(model)
        return model

    def _add_value_types(self, model):
        enum_to_values = defaultdict(list)
        for rvc in self._parser.role_value_constraints().values():
            if len(rvc.values) > 0 and isinstance(rvc.values[0], str):
                enum_to_values[rvc.role.player].extend(rvc.values)
        # create an Enum instead of ValueType when there is a role value constraint (only string sequence) played by this ValueType
        for vt in self._parser.value_types().values():
            if vt.id in enum_to_values:
                model.Enum(to_pascal_case(vt.name), enum_to_values[vt.id])
            else:
                model.Concept(vt.name, extends=[self._datatype_mapping.get(vt.data_type, String)])

    def _add_entity_types(self, model):
        extended_concepts = [y.subtype_name for x in self._parser.subtype_facts().values() for y in x]

        for et in self._parser.entity_types().values():
            if et.name not in extended_concepts:
                model.Concept(et.name)

        subtype_facts = self._parser.sorted_subtype_facts()
        for st_fact in subtype_facts:
            parent = st_fact.supertype_name
            parent_entity = model.lookup_concept(parent)
            if parent_entity is None:
                parent_entity = model.Concept(parent)
            model.Concept(st_fact.subtype_name, extends=[parent_entity])

    def _add_subtype_relationships(self, model):
        for parent, children in self._parser.subtype_facts().items():
            exclusives = []
            exclusives_and_inclusives = []
            for child in children:
                sub = model.lookup_concept(child.subtype_name)
                if isinstance(child, ExclusiveInclusiveSubtypeFact):
                    exclusives_and_inclusives.append(sub)
                elif isinstance(child, ExclusiveSubtypeFact):
                    exclusives.append(sub)

            if len(exclusives_and_inclusives) > 0:
                model.exclusive_subtype_constraint(*exclusives_and_inclusives)
                model.inclusive_subtype_constraint(*exclusives_and_inclusives)
            if len(exclusives) > 0:
                model.exclusive_subtype_constraint(*exclusives)

    def _add_relationships(self, model):
        object_types = self._parser.object_types()
        unique_roles = self._parser.unique_roles()
        mandatory_roles = self._parser.mandatory_roles()
        role_value_constraints = self._parser.role_value_constraints()
        fact_type_to_internal_ucs = self._parser.fact_type_to_internal_ucs()
        fact_type_to_complex_ucs = self._parser.fact_type_to_complex_ucs()
        fact_type_to_roles = self._parser.fact_type_to_roles()
        for fact_type, reading_orders in self._parser.fact_type_readings().items():

            # Adding the main reading
            rdo = reading_orders[0]
            player = object_types[rdo.roles[0].role.player].name
            player_entity = model.lookup_concept(player)
            rel_name, reading = self._build_relation(model, fact_type, player_entity, rdo)
            relationship = model.Relationship(reading)
            setattr(player_entity, rel_name, relationship)
            self._relationships[fact_type] = relationship

            # Marking unique and mandatory roles
            role_idx_to_player = []
            for ro in rdo.roles:
                role_id = ro.role.id
                role_idx_to_player.append(role_id)
                role_index = role_idx_to_player.index(role_id)
                role = relationship[role_index]
                if role_id in unique_roles:
                    model.unique(role)
                if role_id in mandatory_roles:
                    model.mandatory(role)
                if role_id in role_value_constraints.keys():
                    self._relationship_role_value_constraints[role] = role_value_constraints[role_id].values

            # Adding alternative readings
            if len(reading_orders) > 1:
                for rdo in reading_orders[1:]:
                    other_player = object_types[rdo.roles[0].role.player].name
                    other_player_entity = model.lookup_concept(other_player)
                    rel_name, alt_reading = self._build_relation(model, fact_type, other_player_entity, rdo)
                    setattr(other_player_entity, rel_name, relationship.alt(alt_reading))

            # Marking identifying relationships
            if fact_type in fact_type_to_internal_ucs:
                for uc in fact_type_to_internal_ucs[fact_type]:
                    if uc.identifies:
                        player_entity.identify_by(relationship)

            # Adding constraint spanning over multiple roles
            if fact_type in fact_type_to_complex_ucs:
                for uc in fact_type_to_complex_ucs[fact_type]:
                    uc_roles = []
                    for role in fact_type_to_roles[fact_type]:

                        if role.id in uc.roles:
                            p = role.id
                            rl = role_idx_to_player.index(p)
                            uc_roles.append(self._relationships[fact_type][rl])
                    model.unique(*uc_roles)

    def _add_external_identifying_relationships(self, model):
        roles = self._parser.roles()
        fact_types_to_readings = self._parser.fact_type_readings()
        object_types = self._parser.object_types()
        for uc_id, uc in self._parser.external_uniqueness_constraints().items():
            entity = model.lookup_concept(object_types[uc.identifies].name)
            identifying_rel = []
            for ro in uc.roles:
                fact_name = roles[ro].relationship_name
                readings = fact_types_to_readings[fact_name]
                for rdo in readings:
                    if rdo.roles[0].role.player == uc.identifies:
                        for rl in rdo.roles:
                            if rl.role.id == ro:
                                identifying_rel.append(self._relationships[fact_name])
            entity.identify_by(*identifying_rel)

    def _add_role_constraints(self, model):
        roles = self._parser.roles()
        for rc in self._parser.role_constraints():
            constraint_roles = []
            for ro in rc.roles:
                constraint_roles.append(self._role_id_to_role_object(model, roles[ro]))
            if rc.inclusive:
                model.inclusive_roles(*constraint_roles)
            if rc.exclusive:
                model.exclusive_roles(*constraint_roles)

    def _add_ring_constraints(self, model):
        roles = self._parser.roles()
        for rc in self._parser.ring_constraints().values():
            constraint_roles = []
            constraint_types = []
            for ro in rc.roles:
                constraint_roles.append(self._role_id_to_role_object(model, roles[ro]))
            for rt in rc.ring_types:
                constraint_types.append(self._ring_type_mapping.get(rt))
            model.ring(constraint_types, *constraint_roles)

    def _add_role_value_constraints(self, model):
        for rel, values in self._relationship_role_value_constraints.items():
            model.role_value_constraint(rel, values)

    def _build_reading(self, model, reading_order):
        object_types = self._parser.object_types()
        rel_args = []
        for rdo_role in reading_order.roles:
            if rdo_role.prefix is not None:
                rel_args.append(rdo_role.prefix + "-")
            p = model.lookup_concept(object_types[rdo_role.role.player].name)
            rel_args.append(f"{{{p}}}") if rdo_role.role.name == "" else rel_args.append(f"{{{rdo_role.role.name}:{p}}}")
            if rdo_role.postfix is not None:
                rel_args.append("-" + rdo_role.postfix)
            if rdo_role.text is not None:
                rel_args.append(" " + rdo_role.text + " ")
        return ''.join(rel_args).strip()

    def _rename_to_ref_mode(self, model, first_player_entity, fact_type):
        identifier_fact_type_to_entity_type = self._parser.identifier_fact_type_to_entity_type()
        entity_type = identifier_fact_type_to_entity_type.get(fact_type)
        if entity_type:
            player = model.lookup_concept(entity_type.name)
            if first_player_entity == player:
                return entity_type.ref_mode
        return None

    def _build_relation(self, model, fact_type, player_entity, reading_order):
        object_types = self._parser.object_types()
        role_name = ""
        for role in reading_order.roles:
            if object_types[role.role.player].name != player_entity._name:
                role_name = role.role.name
        reading = self._build_reading(model, reading_order)
        ref_mode = self._rename_to_ref_mode(model, player_entity, fact_type)
        if ref_mode is not None:
            rel_name = ref_mode.lower()
        else:
            rel_name = role_name if role_name != "" else build_relation_name_from_reading(reading)
        return rel_name, reading

    def _role_id_to_role_object(self, model, role_obj):
        player_entity = model.lookup_concept(self._parser.object_types()[role_obj.player].name)
        rel = self._relationships[role_obj.relationship_name]
        idx = rel._field_names.index(player_entity._name.lower())
        return rel[idx]



import logging
from typing import Optional

import relationalai.early_access.builder as qb
from relationalai.early_access.dsl.orm.constraints import Unique, Mandatory, RoleValueConstraint, \
    InclusiveSubtypeConstraint, ExclusiveSubtypeConstraint
from relationalai.early_access.dsl.orm.relationships import Role, Relationship
from relationalai.early_access.metamodel.util import OrderedSet


class OntologyReasoner:
    def __init__(self, model):
        self._model = model

        # TODO : populate and use
        self._value_types: OrderedSet[qb.Concept] = OrderedSet()
        self._entity_types: OrderedSet[qb.Concept] = OrderedSet()
        self._exclusive_entity_types: OrderedSet[qb.Concept] = OrderedSet()

        self._concept_identifiers: dict[qb.Concept, OrderedSet[Unique]] = {}
        self._subtype_identifiers: dict[qb.Concept, OrderedSet[Unique]] = {}
        self._constraint_identifies: dict[Unique, qb.Concept] = {}
        self._constructor_roles: OrderedSet[Role] = OrderedSet()
        self._mandatory_roles: OrderedSet[Role] = OrderedSet()

        self._role_value_constraints: OrderedSet[RoleValueConstraint] = OrderedSet()
        self._inclusive_subtype_constraints: OrderedSet[InclusiveSubtypeConstraint] = OrderedSet()
        self._exclusive_subtype_constraints: OrderedSet[ExclusiveSubtypeConstraint] = OrderedSet()

        self._subtype_map: dict[qb.Concept, OrderedSet[qb.Concept]] = {}
        self._supertype_map: dict[qb.Concept, OrderedSet[qb.Concept]] = {}
        self._subtype_closure: dict[qb.Concept, OrderedSet[qb.Concept]] = {}
        self._supertype_closure: dict[qb.Concept, OrderedSet[qb.Concept]] = {}
        self._type_closure: dict[qb.Concept, OrderedSet[qb.Concept]] = {}

        self._analyze()

    #=
    # Constraints: uniqueness and reference schemes.
    #=

    def ref_schemes_of(self, entity_type: qb.Concept, shallow=False) -> OrderedSet[Unique]:
        """
        Returns the reference schemes that are defined for the entity type (plus those inferred from sub- or supertypes).
        """
        own_reference_schemes = self._concept_identifiers.get(entity_type) or OrderedSet()
        if shallow or self.is_exclusive_entity_type(entity_type):
            subtype_reference_schemes = self._subtype_identifiers.get(entity_type) or OrderedSet()
        else:
            subtype_reference_schemes = self.subtype_ref_schemes_of(entity_type)
        return subtype_reference_schemes | own_reference_schemes

    def subtype_unique_ref_scheme_role(self, entity_type: qb.Concept) -> Role:
        """
        Returns the role used to construct the entity type.

        Only valid for simple reference schemes inferred by a subtype.
        """
        ref_schemes = self.ref_schemes_of(entity_type)
        if len(ref_schemes) != 1:
            raise ValueError(f"No unique reference scheme found for {entity_type}")
        ref_scheme = ref_schemes[0]
        return self._extract_unique_role_from_uc(ref_scheme)

    def subtype_ref_schemes_of(self, entity_type: qb.Concept) -> OrderedSet[Unique]:
        """
        Returns the reference schemes that are defined for the entity type and any of its parent types.
        """
        ref_schemes = entity_type._ref_scheme()
        if ref_schemes is None:
            raise ValueError(f"No reference scheme found for {entity_type}")
        identifier_constraints = OrderedSet()
        for relationship in ref_schemes:
            if not isinstance(relationship, Relationship):
                raise ValueError(f"Expected a Relationship, but got {type(relationship)}")
            identifier_constraints.add(self.lookup_identifier_constraint(relationship))
        return identifier_constraints

    def own_ref_scheme_of(self, entity_type: qb.Concept) -> Unique:
        """
        Returns the reference schemes that are defined for the entity type (but not inferred).
        """
        own_ref_schemes = self._concept_identifiers[entity_type]
        if len(own_ref_schemes) != 1:
            raise ValueError(f"Expected exactly one reference scheme for {entity_type}, but got {len(own_ref_schemes)}")
        return own_ref_schemes[0]

    def own_ref_scheme_role(self, entity_type: qb.Concept) -> Role:
        """
        Returns the role used to construct the entity type. Only valid for simple reference schemes.
        """
        ref_scheme = self.own_ref_scheme_of(entity_type)
        return self._extract_unique_role_from_uc(ref_scheme)

    def has_own_ref_scheme(self, entity_type: qb.Concept) -> bool:
        """
        Returns true if the entity type has a reference scheme defined for it.
        """
        return entity_type in self._concept_identifiers

    def has_simple_ref_scheme(self, entity_type: qb.Concept) -> bool:
        ref_schemes = self.ref_schemes_of(entity_type)
        if len(ref_schemes) == 0:
            raise ValueError(f"No reference scheme found for {entity_type}")
        for uc in ref_schemes:
            if len(uc.roles()) > 1:
                return False
        return True

    def has_composite_ref_scheme(self, entity_type: qb.Concept) -> bool:
        return not self.has_simple_ref_scheme(entity_type)

    def is_constructing_role(self, role: Role) -> bool:
        """
        Returns true if the role is a constructing role (i.e., it is used to construct an entity type).
        """
        return role in self._constructor_roles

    def concept_identifiers(self) -> dict[qb.Concept, OrderedSet[Unique]]:
        """
        Returns a mapping of entity types to their reference schemes.
        """
        return self._concept_identifiers

    def is_identifier_relationship(self, relationship: Relationship) -> bool:
        """
        Returns true if the relationship is an identifier relationship (i.e., it is used to construct an entity type).
        """
        if not relationship._binary():
            return False
        for role in relationship._roles():
            if self.is_constructing_role(role):
                return True
        return False

    def lookup_identifier_constraint(self, relationship: Relationship) -> Unique:
        """
        Returns the corresponding identifier constraint for an identifier relationship.
        """
        if not self.is_identifier_relationship(relationship):
            raise ValueError(f"Expected an identifier relationship, but got {relationship}")
        concept_role = None
        constructing_role = None
        for role in relationship._roles():
            if self.is_constructing_role(role):
                concept_role = role.sibling()
                constructing_role = role
            else:
                constructing_role = role.sibling()
                concept_role = role
        assert concept_role is not None and constructing_role is not None, \
            f"Expected both roles to be defined in {relationship}"
        for constraint in self._concept_identifiers[concept_role.player()]:
            if constructing_role in constraint.roles():
                return constraint
        raise ValueError(f"No identifier constraint found for {relationship}")

    def identifies_concept(self, constraint: Unique) -> qb.Concept:
        """
        Returns the concept that is identified by the given uniqueness constraint.
        """
        if constraint not in self._constraint_identifies:
            raise KeyError(f"No concept identified by {constraint}")
        return self._constraint_identifies[constraint]

    def is_composite_concept(self, concept: qb.Concept) -> bool:
        """
        Returns true if the concept is a composite concept (i.e., it has a composite reference scheme).
        """
        ref_schemes = self.ref_schemes_of(concept)
        if len(ref_schemes) == 0:
            return False
        for uc in ref_schemes:
            if len(uc.roles()) > 1:
                return True
        return False

    def is_exclusive_entity_type(self, entity_type: qb.Concept) -> bool:
        """
        Returns true if the entity type is an exclusive entity type).
        """
        return entity_type in self._exclusive_entity_types or self._check_exclusive_supertype(entity_type)

    def _check_exclusive_supertype(self, entity_type: qb.Concept) -> bool:
        """
        Checks if the entity type is an exclusive supertype.
        """
        if self.has_own_ref_scheme(entity_type):
            return False
        subtypes = self._subtype_map.get(entity_type)
        is_exclusive = subtypes is not None
        if is_exclusive:
            self._exclusive_entity_types.add(entity_type)
        return is_exclusive

    def subtype_exclusive_supertype(self, subtype: qb.Concept) -> Optional[qb.Concept]:
        """
        Returns the exclusive supertype of the given subtype if it exists.
        """
        supertypes = self._supertype_map.get(subtype)
        if not supertypes:
            return None
        for supertype in supertypes:
            if self.is_exclusive_entity_type(supertype):
                return supertype
        return None

    @staticmethod
    def _extract_unique_role_from_uc(uc: Unique) -> Role:
        """
        Extracts the unique role from a uniqueness constraint.
        """
        roles = uc.roles()
        if len(roles) != 1:
            raise ValueError(f"Expected exactly one role in uniqueness constraint, but got {len(roles)}")
        return roles[0]

    #=
    # Analysis based on the model.
    #=

    def _analyze(self):
        self._analyze_constraints()
        self._analyze_subtypes()

    #=
    # Look through the constraints, capture the ones that are reference schemes.
    #=

    def _analyze_constraints(self):
        for constraint in self._model.constraints():
            if isinstance(constraint, Unique):
                self._process_uniqueness_constraint(constraint)
            elif isinstance(constraint, Mandatory):
                self._process_mandatory_constraint(constraint)
            elif isinstance(constraint, RoleValueConstraint):
                self._role_value_constraints.add(constraint)
            elif isinstance(constraint, InclusiveSubtypeConstraint):
                self._inclusive_subtype_constraints.add(constraint)
            elif isinstance(constraint, ExclusiveSubtypeConstraint):
                self._exclusive_subtype_constraints.add(constraint)
            else:
                logging.warning(f"Unknown constraint type: {type(constraint)}")

    def _process_uniqueness_constraint(self, constraint: Unique):
        # note: ignoring the ones that are not preferred identifiers for now
        if not constraint.is_preferred_identifier:
            return

        roles = constraint.roles()
        self._constructor_roles.update(roles)  # mark as ctor roles

        if constraint._is_internal():
            # simple ref scheme
            self._process_simple_ref_scheme(constraint)
        else:
            # composite ref scheme
            self._process_composite_ref_scheme(constraint)

    def _process_simple_ref_scheme(self, constraint: Unique):
        role = constraint.roles()[0]
        relation = role._relationship  # actually a DSL Relation
        if relation._arity() != 2:
            raise ValueError(f"Identifier relationship {relation} should have arity 2, but got {relation._arity()}")

        sibling_role = role.sibling()
        assert sibling_role is not None, f"Unable to find the sibling role for {role}"
        constructed_concept = sibling_role.player()
        self._concept_identifiers.setdefault(constructed_concept, OrderedSet()).add(constraint)
        self._constraint_identifies[constraint] = constructed_concept

    def _process_composite_ref_scheme(self, constraint: Unique):
        roles = constraint.roles()
        if len(roles) <= 1:
            raise ValueError(f"External uniqueness constraint should have more than one role, but got {len(roles)}")
        role = roles[0]  # take an arbitrary role to get the constructed concept
        sibling_role = role.sibling()
        assert sibling_role is not None, f"Unable to find the sibling role for {role}"
        constructed_concept = sibling_role.player()
        self._concept_identifiers.setdefault(constructed_concept, OrderedSet()).add(constraint)
        self._constraint_identifies[constraint] = constructed_concept

    def _process_mandatory_constraint(self, constraint: Mandatory):
        role = constraint.roles()[0]
        self._mandatory_roles.add(role)

    #=
    # Look through the subtype hierarchies and capture the preferred IDs (inferring as necessary).
    #=

    def _analyze_subtypes(self):
        self._initialize_type_hierarchy()
        self._infer_subtype_identifiers()
        self._compute_type_closures()

    def _initialize_type_hierarchy(self):
        for child_concept in self._model.entity_types():
            for parent_concept in child_concept._extends:
                self._subtype_map.setdefault(parent_concept, OrderedSet()).add(child_concept)
                self._supertype_map.setdefault(child_concept, OrderedSet()).add(parent_concept)

    def _infer_subtype_identifiers(self):
        (leaf_types, non_leaf_types) = self._partition_type_hierarchy()

        # if a leaf has a ref scheme, then that's what we use; else look it up bottom-up and throw if none found
        memo = set()
        for leaf_type in leaf_types:
            self._find_supertype_identifier_bottom_up(leaf_type, memo, [])

        #=
        # For any intermediate or root nodes that didn't get a ref scheme during the bottom-up lookup, try finding one
        # by doing a top-down lookup (may result in multiple ref schemes from different subtypes).
        #=
        for non_leaf_type in non_leaf_types:
            if non_leaf_type in memo:
                continue
            try:
                self._find_supertype_identifier_bottom_up(non_leaf_type, memo, [])
            except KeyError:
                # mark concept as an exclusive entity type and attempt to find subtype identifiers
                self._exclusive_entity_types.add(non_leaf_type)
                # ignore error, since the top-down approach will throw if none exists either
                self._find_subtype_identifiers_top_down(non_leaf_type, memo, [])

    def _partition_type_hierarchy(self):
        leaf_types = OrderedSet()
        non_leaf_types = OrderedSet()
        non_leaf_types.update(self._subtype_map.keys())
        for entity in self._supertype_map:
            if entity in self._subtype_map:
                non_leaf_types.add(entity)
            else:
                leaf_types.add(entity)
        return leaf_types, non_leaf_types

    def _find_supertype_identifier_bottom_up(self, entity_type: qb.Concept, memo: set[qb.Concept],
                                             lookup_chain: list[qb.Concept]):
        """
        Looks up a ref scheme for an entity type bottom up or throws an error otherwise.

        It's designed to be used on leaf types. If used on intermediate nodes, if they don't have a ref scheme, any
        exception should be swallowed and top-down used to look up a set of possible ref schemes coming from its
        subtypes.
        """
        if entity_type in memo:
            return self._subtype_identifiers.get(entity_type)

        if lookup_chain is None:
            lookup_chain = []
        lookup_chain.append(entity_type)

        ref_schemes = self._concept_identifiers.get(entity_type)
        if ref_schemes is not None:
            self._subtype_identifiers.setdefault(entity_type, OrderedSet()).update(ref_schemes)
            memo.add(entity_type)  # use memoization to reduce lookups
            return ref_schemes
        # Check supertype
        supertypes = self._supertype_map.get(entity_type)
        if supertypes is None:
            raise KeyError(f'No reference scheme found for {entity_type}, in chain {lookup_chain}')
        if len(supertypes) > 1:
            raise NotImplementedError(f'Multiple supertypes are not supported yet (in {entity_type})')
        for supertype in supertypes:
            ref_schemes = self._find_supertype_identifier_bottom_up(supertype, memo, lookup_chain)
            self._subtype_identifiers.setdefault(entity_type, OrderedSet()).update(ref_schemes)
            memo.add(entity_type)  # use memoization to reduce lookups
            return ref_schemes
        raise KeyError(f'No reference scheme found for {entity_type}, in chain {lookup_chain}')

    def _find_subtype_identifiers_top_down(self, entity_type: qb.Concept, memo: set[qb.Concept],
                                           lookup_chain: list[qb.Concept]):
        """
        Looks up a ref scheme for an entity type top-down or throws an error otherwise.

        May result in multiple ref schemes from different subtypes (branches in the subtype tree).
        """
        if entity_type in memo:
            return self._subtype_identifiers.get(entity_type)

        subtypes = self._subtype_map.get(entity_type)
        if subtypes is None:
            raise KeyError(f'No reference scheme found for {entity_type}, in chain {lookup_chain}')

        ref_schemes = OrderedSet()
        for subtype in subtypes:
            subtype_ref_schemes = self._concept_identifiers.get(subtype)
            if subtype_ref_schemes is not None:
                ref_schemes.update(subtype_ref_schemes)
            else:
                lookup_chain.append(subtype)
                subtype_ref_schemes = self._find_subtype_identifiers_top_down(subtype, memo, lookup_chain)
                ref_schemes.update(subtype_ref_schemes)

        self._subtype_identifiers.setdefault(entity_type, OrderedSet()).update(ref_schemes)
        memo.add(entity_type)  # use memoization to reduce lookups
        return ref_schemes

    def _compute_type_closures(self):
        entity_types = self._subtype_map.keys() | self._supertype_map.keys()
        for entity_type in entity_types:
            self._subtype_closure[entity_type] = self._compute_closure(entity_type, self._subtype_map)
            self._supertype_closure[entity_type] = self._compute_closure(entity_type, self._supertype_map)
            self._type_closure[entity_type] = self._subtype_closure[entity_type] | self._supertype_closure[entity_type]

    @staticmethod
    def _compute_closure(entity_type: qb.Concept, closure_map: dict[qb.Concept, OrderedSet[qb.Concept]]) -> OrderedSet[qb.Concept]:
        closure = OrderedSet()
        stack = [entity_type]
        while stack:
            current = stack.pop()
            for related in closure_map.get(current, []):
                if related not in closure:
                    closure.add(related)
                    stack.append(related)
        return closure



from relationalai.early_access.metamodel import ir as meta
from relationalai.early_access.metamodel import types
from relationalai.early_access.lqp import ir as lqp

def meta_type_to_lqp(typ: meta.Type) -> lqp.RelType:
    if isinstance(typ, meta.UnionType):
        # By this point, unions can only consist of entity types. In the LQP,
        # they are undifferentiated (hashes), so they all merge.
        assert all(types.is_entity_type(t) for t in typ.types)
        return lqp.PrimitiveType.UINT128
    else:
        assert isinstance(typ, meta.ScalarType)
        if types.is_builtin(typ):
            # TODO: just ocompare to types.py
            if typ.name == "Int":
                return lqp.PrimitiveType.INT
            elif typ.name == "Float":
                return lqp.PrimitiveType.FLOAT
            elif typ.name == "String":
                return lqp.PrimitiveType.STRING
            elif typ.name == "Decimal64":
                return lqp.RelValueType.DECIMAL64
            elif typ.name == "Decimal128":
                return lqp.RelValueType.DECIMAL128
            elif typ.name == "Date":
                return lqp.RelValueType.DATE
            elif typ.name == "DateTime":
                return lqp.RelValueType.DATETIME
            elif typ.name == "RowId":
                return lqp.PrimitiveType.UINT128
            elif typ.name == "Number":
                # All types must be specified in the LQP.
                raise Exception("Number type could not be determined.")
            elif types.is_any(typ):
                # All types must be specified in the LQP.
                raise Exception("Type could not be determined.")
            else:
                raise NotImplementedError(f"Unknown builtin type: {typ.name}")
        elif types.is_entity_type(typ):
            return lqp.PrimitiveType.UINT128
        else:
            # Otherwise, the type extends some other type, we use that instead
            assert len(typ.super_types) > 0, f"Type {typ} has no super types"
            assert len(typ.super_types) == 1, f"Type {typ} has multiple super types: {typ.super_types}"
            super_type = typ.super_types[0]
            assert isinstance(super_type, meta.ScalarType), f"Super type {super_type} of {typ} is not a scalar type"
            return meta_type_to_lqp(super_type)

def type_from_constant(arg: lqp.PrimitiveValue) -> lqp.RelType:
    if isinstance(arg, int):
        return lqp.PrimitiveType.INT
    elif isinstance(arg, float):
        return lqp.PrimitiveType.FLOAT
    elif isinstance(arg, str):
        return lqp.PrimitiveType.STRING
    elif isinstance(arg, lqp.UInt128):
        return lqp.PrimitiveType.UINT128
    else:
        raise NotImplementedError(f"Unknown constant type: {type(arg)}")



from relationalai.early_access.dsl.core.constraints import Constraint
from relationalai.early_access.dsl.core.constraints.scalar import ScalarConstraint
from relationalai.early_access.dsl.core.exprs import Expr, Wildcard

# An AtomicConstraint is a Constraint that will ultimately realize as a
# literal in the head or body of a datalog rule. The concrete subclass
# Atom allows users to declare these directly, while other subclasses
# provide a slightly higher-level means for doing so by treating
# relations as sets.
#
class AtomicConstraint(Constraint):

    relcomp = None

    def component(self): return self.relcomp # type: ignore

    def decorated(self) -> bool:
        return self.relcomp.decorated() # type: ignore

    def undecorated(self) -> bool:
        return self.relcomp.undecorated() # type: ignore


# An Atom is a constraint that can be used as a literal in the
# head or body of a datalog rule, i.e.:
#
#   R(x1, x2, ..., xn)
#
# or:
#
#   not R(x1, x2, ..., xn)
#
# where x1, x2, ..., xn are arithmetic expressions.
#
class Atom(AtomicConstraint):

    def __init__(self, r, args):
        self.relcomp = r
        self.args = args
        self.negated = False

    def entityid(self):
        hvals = [a.entityid() for a in self.args]
        hvals.append(self.relcomp.entityid())
        return hash(tuple(hvals))

    def grounds(self, v, groundings):
        for a in self.args:
            if a.entityid() == v.entityid():
                return True
        return False

    # Returns a new Atom that duplicates self after mapping
    # the Relation component through relmap
    #
    def map(self, relmap):
        newrel = relmap[self.relcomp.entityid()]
        newatom = Atom(newrel, self.args)
        newatom.negated = self.negated
        return newatom

    # Returns a new Atom that is constructed by the given
    # builder function
    #
    def map_builder(self, builder):
        new_atom = builder(self.relcomp, self.args)
        new_atom.negated = self.negated
        return new_atom

    def __invert__(self):
        self.negated = not self.negated
        return self

    def relation(self):
        return self.relcomp.typeof() if hasattr(self.relcomp, 'typeof') else self.relcomp

    # Emit this schema to a textual output using the Z display style
    #
    def pprint(self):
        relname = self.relcomp.display()
        head_str = ", ".join([v.display() for v in self.args])
        patom = f"{relname}({head_str})"
        if self.negated:
            return "not " + patom
        else:
            return patom

    def rel_formula(self):
        relname = self.relcomp.display()
        head_str = ", ".join([v.display() for v in self.args])

        if self.relation().arity() == len(self.args):
            patom = f"{relname}({head_str})"
        else:
            patom = f"{relname}[{head_str}]"

        if self.negated:
            return "not " + patom
        else:
            return patom

    def scalar_refs(self):
        dic = {}
        for a in self.args:
            if a.variable():
                dic[a.display()] = a
            else:
                cdic = a.scalar_refs()
                for v in cdic:
                    if v not in dic:
                        dic[v] = cdic[v]
        return dic

    def refersto(self, varname: str) -> bool:
        if self.relcomp.refersto(varname):
            return True
        for a in self.args:
            if a.refersto(varname):
                return True
        return False

    def rename(self, renaming):
        #
        # `self.relcomp` can either be a `RelationalComponent` or another type such as `ExternalRelation`.
        # Note that `RelationalComponent` and `ExternalRelation` do not have a common supertype.
        # Only instances of `RelationalComponent` support renaming, so we need to check for that here.
        # We need to import `RelationalComponent` locally to avoid circular dependencies.
        #
        # TODO: This requires a proper fix, which would possibly introduce a common supertype for the types which can
        #  occur here and separate the relation and variable renamings into distinct maps (as is done elsewhere, e.g.,
        #  in callers of this method).
        #
        from relationalai.early_access.dsl.schemas import RelationalComponent
        new_relcomp = self.relcomp.rename(renaming) if isinstance(self.relcomp, RelationalComponent) else self.relcomp
        renamed = Atom(new_relcomp,
                       [a.rename(renaming) for a in self.args])
        renamed.negated = self.negated
        return renamed

    def simplify(self):
        return self

    def substitute(self, bindings):
        return Atom(self.relcomp.substitute(bindings),
                    [a.substitute(bindings)
                     for a in self.args])

    # [REKS] Need a better name. The purpose of this method is to undash an atom
    #        that involves relation R when we are generating a recursive Rel rule
    #        from an operation Schema involving relational components that are
    #        dashed and undashed occurrences of R. The "undashed" atom will be
    #        used to form the head of this rule.
    #
    def rule_head(self):
        return Atom(self.relcomp.undash(), self.args)

    def value_column(self):
        rel = self.relation()
        pos = rel.arity() - 1  # Position of value component
        return self.args[pos]


class ElementOf(AtomicConstraint):

    def __init__(self, x, y):
        self.element = x
        self.relcomp = y

        # Emit this schema to a textual output using the Z display style

    #
    def pprint(self):
        return self.element.display() + " IN " + self.relcomp.display()

    def rel_formula(self):
        elem = self.element
        if elem.relational():
            raise Exception(f"Cannot make expr {elem.display()} an argument of an AtomicConstraint")

        return self.relcomp.project(elem).rel_formula()

    def grounded(self):
        return self.element.grounded()

    def grounds(self, v, groundings):
        return self.element.entityid() == v.entityid()

    def scalar_refs(self):
        dic = self.relcomp.scalar_refs()
        elem = self.element
        if elem.variable():
            dic[elem.display()] = elem
        else:
            edic = elem.scalar_refs()
            for v in edic:
                if v not in dic:
                    dic[v] = edic[v]
        return dic

    def refersto(self, varname: str) -> bool:
        return self.element.refersto(varname) or self.relcomp.refersto(varname)

    def rename(self, renaming):
        return ElementOf(self.element.rename(renaming),
                         self.relcomp.rename(renaming))

    def simplify(self):
        return ElementOf(self.element, self.relcomp.simplify())

    def substitute(self, bindings):
        e = self.element.substitute(bindings)
        p = self.relcomp.substitute(bindings)
        return ElementOf(e, p)

    def entityid(self):
        return hash((ElementOf,
                     self.element.entityid(),
                     self.relcomp.entityid()))


class NotElementOf(AtomicConstraint):

    def __init__(self, x, y):
        self.element = x
        self.relcomp = y

    # Emit this schema to a textual output using the Z display style
    #
    def pprint(self):
        return self.element.display() + " NOT IN " + self.relcomp.display()

    def rel_formula(self):
        return "not " + self.relcomp.project(self.element).rel_formula()

    def grounded(self):
        return self.element.grounded()

    def grounds(self, v, groundings):
        return False

    def scalar_refs(self):
        dic = self.relcomp.scalar_refs()
        elem = self.element
        if elem.variable():
            dic[elem.display()] = elem
        else:
            edic = elem.scalar_refs()
            for v in edic:
                if v not in dic:
                    dic[v] = edic[v]
        return dic

    def refersto(self, varname: str) -> bool:
        return self.element.refersto(varname) or self.relcomp.refersto(varname)

    def rename(self, renaming):
        return NotElementOf(self.element.rename(renaming),
                            self.relcomp.rename(renaming))

    def simplify(self):
        return ElementOf(self.element, self.relcomp.simplify())

    def substitute(self, bindings):
        e = self.element.substitute(bindings)
        p = self.relcomp.substitute(bindings)
        return NotElementOf(e, p)

    def entityid(self):
        return hash((NotElementOf,
                     self.element.entityid(),
                     self.relcomp.entityid()))


# This module declares operations that range over sets of scalar and
# relational constraints.
#
# By convention, we model constraint sets as dictionaries that map
# a hash of the Constraint to the Constraint itself to aid in
# deduplication.
#


# AtomicConstraint set optimizations follow


# Given a set A of AtomicConstraints:
#
#   - Remove any obviously redundant atoms, and
#   - Unify any pair { c1, c2 } subseteq A, if c1 and c2 refer to the
#     same Relation and that Relation is functional, then if c2 and c2
#     can be shown to refer to the same fact, add new equality
#     constraints to record this knowledge.
#
# Returns the set A less any redundant atoms and extended by any equality
# constraints that can be discovered through unification.
#
# We peform unification by constructing a UnificationTree (Utree), which is
# a recursive dictionary of dictionaries ultimately terminating in lists of
# atoms from A. Each dictionary at depth k implements the decision to extend
# the tuple formed by decisions made at depths > k with a decision (column
# value) at depth k. For instance, were A to contain the following atoms:
#
#    a1 = foo(x, y, 2)
#    a2 = foo(x, y, v)
#    a3 = foo(a, y, c)
#    a4 = foo(a, z, d)
#
# then the Utree would look like this:
#
#    { a: { y: [a3],
#           z: [a4] },
#      x: { y: [a1, a2] } }
#
# where {a, x, y, z} in these keys are the hashes of those expressions
#
def reduce_relational_atoms(A):

    # Given a set of AtomicConstraints A and a Relation hash r, returns those
    # atoms in A that reference the Relation whose hash is r
    #
    def references(A, r):
        return {a.entityid(): a
                for a in A.values()
                if a.relcomp.entityid() == r}

    # Given a set of AtomicConstraints A, return the minimal set of relations
    # such that each Relation in the return set is referenced in at least
    # one atom in A
    #
    def rel_refs_of(A):
        rels = {}
        for a in A.values():
            rel = a.relcomp
            rv = rel.entityid()
            if rv not in rels:
                rels[rv] = rel
        return rels

    # A UnificationTree (Utree) is a decision-tree representation (ordered
    # by column number) of one or more facts that are asserted by a set of
    # AtomicConstraints that all reference the same Relation
    #
    def utree_digest(utree, depth, args, atom):
        cval = args[0]
        v = cval.entityid() if isinstance(cval, Expr) else cval
        if v in utree:
            if depth == 1:
                utree[v].append(atom)
            else:
                utree_digest(utree[v], depth - 1, args[1:], atom)
        else:
            if depth == 1:
                utree[v] = [atom]
            else:
                utree[v] = {}
                utree_digest(utree[v], depth - 1, args[1:], atom)

    # Flatten a Utree by recording all of the atoms at its leaves
    # in the set atoms
    #
    def utree_export(utree, depth, atoms):
        for v in utree.values():
            if depth == 1:
                atoms[a.entityid()] = a
            else:
                utree_export(v, depth - 1, atoms)
        return atoms

    # Mutate utree to remove atoms with Wildcard values when we can show
    # those atoms are redundant
    #
    def utree_remove_wildcards(utree, depth):
        if depth == 1:
            for n in utree.values():
                if len(n) > 1:
                    # If a functional relation has an n-1 tuple prefix that maps
                    # to multiple values, then any atom with a Wildcard value
                    # is redundant and can be removed
                    #
                    for a in n:
                        if isinstance(a.value_column(), Wildcard):
                            n.remove(a)
        else:
            for pref in utree:
                utree_remove_wildcards(utree[pref], depth - 1)

    # Given a Utree, whose paths represent a set of referent_atoms that refer to some
    # functional relation, walk that tree to unify referent_atoms that differ only in
    # the last column so as to produce two sets:
    #
    #  - unique_atoms, which is a minimal set of referent atoms such that no two
    #    atoms will reference the same key tuple; and
    #
    #  - unifying_constraints, which is a set of equality constraints that equate
    #    any expressions that that arise from unifying two or more paths in the tree
    #    that reference the same key tuple.
    #
    # This function should be called with a root Utree, its depth, and two empty
    # dictionaries that represent the sets unique_atoms and unifying_constraints,
    # and the function will recursively walk the tree to fill in these sets.
    #
    # If the set referent_atoms already has the unique key tuple property, then on
    # return:
    #
    #    unique_atoms == referent_atoms and unifying_constraints == {}
    #
    # Otherwise, on return:
    #
    #    unique_atoms \subset referent_atoms and unifying_constraints != {}
    #
    def utree_unify_and_export(utree, depth, unique_atoms, unifying_constraints):
        for v in utree.values():
            if depth == 1:
                prev_value = None
                for idx, a in enumerate(v):
                    last_col = a.value_column()
                    if idx == 0:
                        prev_value = last_col
                        unique_atoms[a.entityid()] = a
                    else:
                        eq = ScalarConstraint(prev_value, "=", last_col)
                        unifying_constraints[eq.entityid()] = eq
            else:
                utree_unify_and_export(v, depth - 1, unique_atoms, unifying_constraints)

    equality_constraints = {}
    atoms = {}
    relations = rel_refs_of(A)
    for r in relations:

        # Depth to which we need to build interior nodes (lists) in the
        # unification tree
        #
        rel = relations[r].typeof()
        depth = rel.arity() - 1

        referents = references(A, r)

        if depth == 0:
            for a in referents:
                atoms[a] = referents[a]
        else:
            # Assemble a Utree for the atoms that reference Relation r
            #
            r_utree = {}
            for a in referents.values():
                utree_digest(r_utree, depth, a.args, a)

            utree_remove_wildcards(r_utree, depth)
            if rel.functional:
                utree_unify_and_export(r_utree, depth, atoms, equality_constraints)
            else:
                utree_export(r_utree, depth, atoms)

    return (equality_constraints, atoms)
